/*
    global
        HomegearWS
        i18n
        interface_mount
        modal_mount
        states_flag_dirty
        toast
*/
/*
    exported
        date_format
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function readCookie(key) {
    const regex = new RegExp('(?:^|; )' + encodeURIComponent(key) + '=([^;]*)');
    const result = regex.exec(document.cookie);

    return result ? result[1] : null;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function homegearRandomUserName() {
    const possible = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
    let id = 'shif-';

    for (let i = 0; i < 8; i++)
        id += possible.charAt(Math.floor(Math.random() * possible.length));

    return id;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function homegear_websocket_security() {
    const opts = interfaceData.options;

    if (location.protocol === 'https:')
        return true;

    if (opts.websocket_user || opts.websocket_password)
        return true;

    if (opts.websocket_security_ssl === undefined ||
        opts.websocket_security_ssl === 'location.protocol')
        return location.protocol === 'https:';

    return !! opts.websocket_security_ssl;

}

function homegear_new() {
    const host = (interfaceData.options.websocket_url == 'location.hostname')
                    ? location.hostname
                    : interfaceData.options.websocket_url;
    const port = (interfaceData.options.websocket_port == 'location.port')
                    ? location.port
                    : Number(interfaceData.options.websocket_port);

    const secure = homegear_websocket_security();

    return new HomegearWS(
        host,
        port,
        homegearRandomUserName(),
        secure,
        ...arguments
    );
}

let homegear = interfaceData.options.websocket_user &&
               interfaceData.options.websocket_password
                ? homegear_new(interfaceData.options.websocket_user,
                               interfaceData.options.websocket_password)
                : homegear_new(readCookie('PHPSESSIDUI'));
////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
homegear.ready(function() {
    var addDevicesAsHomegearPeers = Object.keys(interfaceData.map_invoke).map(Number);

    console.log('HomegearPeers: ');
    console.log(addDevicesAsHomegearPeers);

    homegear.addPeers( addDevicesAsHomegearPeers );
    $('#loadingPage').hide();
});

homegear.connected(function() {

});

homegear.disconnected(function() {
    $('#loadingPage').show();
});

////////////////////////////////////////////////////////////////////////////////////////////////////////
//get new sessionid when websocket reconnects
////////////////////////////////////////////////////////////////////////////////////////////////////////
homegear.reconnected(function() {
    window.location.reload(true);
});

////////////////////////////////////////////////////////////////////////////////////////////////////////
//print Homegear errors
////////////////////////////////////////////////////////////////////////////////////////////////////////
homegear.error(function (message) {
    console.log(message);
});

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
if(location.protocol == 'https:' && interfaceData.options.websocket_security_ssl == false)
    alert('Error: If you connect to the interface via https you have to use a secure websocket connection!');
else homegear.connect();

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function roles_relevant(roles) {
    return roles.filter(x => x.hasOwnProperty('id') && x.direction !== 1);
}




function handle_update_event(resp) {
    // DEBUG: works as expected
    const peer    = resp.params[1],
          channel = resp.params[2],
          name    = resp.params[3];
    let   value   = resp.params[4];

    if (!(peer    in interfaceData.map_invoke &&
          channel in interfaceData.map_invoke[peer] &&
          name    in interfaceData.map_invoke[peer][channel]))
        return;

    const inputs = interfaceData.map_invoke[peer][channel][name];
    for (const input of inputs) {
        if (!(input.databaseId in interfaceData.devices &&
              input.control    in interfaceData.devices[input.databaseId]
                                               .controls &&
              input.input      in interfaceData.devices[input.databaseId]
                                               .controls[input.control]
                                               .variableInputs))
            continue;

        if (input.roles === undefined)
            continue;

        interfaceData.devices[input.databaseId]
            .controls[input.control]
            .variableInputs[input.input].properties.value = value;

        states_flag_dirty(roles_relevant(input.roles).map(x => x.id));
    }
}



function date_format() {
    const dt = new Date ();

    const day    = dt.getDate().toString().padStart(2, '0');
    const month  = (dt.getMonth()+1).toString().padStart(2, '0');
    const year   = dt.getFullYear().toString().padStart(4, '0');
    const hour   = dt.getHours().toString().padStart(2, '0');
    const minute = dt.getMinutes().toString().padStart(2, '0');
    const second = dt.getSeconds().toString().padStart(2, '0');

    return `${day}.${month}.${year} ${hour}:${minute}:${second}`;
}



function handle_update_request_ui_refresh(_resp) {
    toast.set({
        close: false,
        content: i18n('refresh.message'),
        buttons: [
            {
                id: 0,
                type: 'success',
                reloadUi: true,
                closeModal: false,
                label: i18n('refresh.message.button.text'),
                icon: 'abort_1',
            },
        ],
    });
}



function handle_update_variable_profile_state_changed(resp) {
    const id    = resp.params[0],
          state = resp.params[1];

    if (interfaceData.profiles[id] === undefined)
        return;

    interfaceData.profiles[id].isActive = state;
}



function handle_update_notification_created(resp) {
    const id = resp.params[0];

    homegear.invoke({
        jsonrpc: '2.0',
        method: 'getUiNotification',
        params: [id, interfaceData.options.language],
    }, function (resp) {
        const new_ = resp.result;

        new_.id = id;

        Vue.set(interfaceData.notifications, id, new_);
    });
}



function handle_update_notification_removed(resp) {
    const id = resp.params[0];

    if (interfaceData.notifications[id] === undefined)
        return;

    Vue.delete(interfaceData.notifications, id);
}



function homegear_handle_update(resp) {
    console.log(JSON.stringify(resp, null, 4));

    const funcs = {
        'event':                       handle_update_event,
        'requestUiRefresh':            handle_update_request_ui_refresh,
        'variableProfileStateChanged': handle_update_variable_profile_state_changed,
        'uiNotificationCreated':       handle_update_notification_created,
        'uiNotificationRemoved':       handle_update_notification_removed,
    };

    if (resp.method in funcs)
        funcs[resp.method](resp);
}



////////////////////////////////////////////////////////////////////////////////////////////////////////
// Extensions to the homegear object
////////////////////////////////////////////////////////////////////////////////////////////////////////
function params_create(input, value) {
    if ('value' in input)
        value = input.value;

    return [
        Number(input.peer),
        Number(input.channel),
        input.name,
        value,
        false,
    ];
}



function homegear_prepare(homegear) {
    homegear.event(homegear_handle_update);
    homegear.ready(interface_mount);
    homegear.ready(modal_mount);

    homegear.invoke_raw = homegear.invoke;
    homegear.invoke = function (op, cb) {
        homegear.invoke_raw(op, function (ret) {
            if (ret.error === undefined)
                return cb ? cb(ret) : undefined;

            console.log('Invoke Error: ' + JSON.stringify(ret.error, null, 4));
        });
    };

    homegear.invoke_multi = function (ops, cb) {
        const object = {
            jsonrpc: '2.0',
            method: 'system.multicall',
            params: ops,
        };

        console.log(JSON.stringify(object, null, 4));

        return this.invoke(object, cb);
    };

    homegear.value_set_multi = function (ops, cb) {
        return this.invoke_multi([
            ops.map(op => ({
                methodName: 'setValue',
                params: params_create(op.input, op.value),
            }))
        ], cb);
    };

    homegear.value_set_clickcounter = function(control, params, value) {
        let methods = [[
            {
                methodName: 'setValue',
                params: params_create(params, value)
            }
        ]];

        if(Date.now() - control.lastClickCount > 10000) {
            control.lastClickCount = Date.now();
            methods[0].push({
                methodName: 'setUserData',
                params: ['ui.clickCounts', control.uiElement.databaseId.toString(), ++control.uiElement.clickCount]
            });
        }

        return this.invoke_multi(methods);
    };
}
homegear_prepare(homegear);

/*
    global
        app
*/
/*
    exported
        ModeMenuState
        mixin_modemenu
*/



const ModeMenuState =  {
    HIDDEN:   -1,
    FAVORITES: 0,
    PROFILES:  1,
    DRAGGABLE: 2,
};



const _modemenu_fields = [
    {
        obj: () => app.favorites,
        field: 'enabled',
    },
    {
        obj: () => app.profiles,
        field: 'enabled',
    },
    {
        obj: () => app.draggable,
        field: 'enabled',
    },
];

let _modemenu_state = Vue.observable({state: ModeMenuState.HIDDEN});



const mixin_modemenu = {
    computed: {
        modemenu_is_hidden: function () {
            return _modemenu_state.state === ModeMenuState.HIDDEN;
        },

        modemenu_is_shown: function () {
            return ! this.modemenu_is_hidden;
        },

        modemenu_state: function () {
            return _modemenu_state.state;
        }
    },

    methods: {
        _modemenu_get_state: function (state) {
            switch (typeof state) {
                case 'string':
                    return ModeMenuState[state];
                case 'number':
                    return state;
            }

            throw 'BUG: state unknown: ' + state;
        },

        modemenu_hide: function () {
            for (const i of _modemenu_fields)
                Vue.set(i.obj(), i.field, false);

            _modemenu_state.state = ModeMenuState.HIDDEN;
        },

        modemenu_show: function(state_) {
            const state = this._modemenu_get_state(state_);

            if (state < ModeMenuState.HIDDEN || state >= _modemenu_fields.length)
                throw 'BUG: state not in MODEMENU: ' + state;

            for (let i = 0; i < _modemenu_fields.length; ++i)
                Vue.set(
                    _modemenu_fields[i].obj(),
                    _modemenu_fields[i].field,
                    state === i
                );

            _modemenu_state.state = state;
        },

        modemenu_is_state: function (state) {
            return _modemenu_state.state === this._modemenu_get_state(state);
        },
    },
};

/*
    global
        mixin_modemenu
        check_disabled_backend
        check_disabled_frontend
*/
/*
    exported
        clone
        set_or_extend
        shif_device
        shif_comps_create
        shif_register_disable_hooks
        user_logoff
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
// globale Variablen
////////////////////////////////////////////////////////////////////////////////////////////////////////
var logFrontend = '';
////////////////////////////////////////////////////////////////////////////////////////////////////////
// write console logs to setting/about/nameClick/log page
////////////////////////////////////////////////////////////////////////////////////////////////////////
if (typeof interfaceData.options.consoleLog !== 'undefined' && (interfaceData.options.consoleLog == 'true' || interfaceData.options.consoleLog == true)) {
    console.oldLog = console.log;
    console.log = function(value) {
        console.oldLog(value);
        viewLog(value);
    };

    console.oldWarn = console.warn;
    console.warn = function(value) {
        console.oldWarn(value);
        viewLog(value);
    };

    console.oldError = console.error;
    console.error = function(value) {
        console.oldError(value);
        viewLog(value);
    };

    console.oldInfo = console.info;
    console.info = function(value) {
        console.oldInfo(value);
        viewLog(value);
    };
}
else {
    console.oldLog = console.log;
    console.log = function(){};
}

function viewLog(value) {
    if (value !== null && typeof value === 'object')
        value = JSON.stringify(value);

    logFrontend += value + '\r\n';

    if($('#log').length){
        $('#log').html(logFrontend);
        $('#log').scrollTop(9999999999);
    }
}


////////////////////////////////////////////////////////////////////////////////////////////////////////
// verhindert, dass der PullFromTopToRefresh im Android Chrome Browser ausgeführt werden kann
////////////////////////////////////////////////////////////////////////////////////////////////////////
let lastTouchY = 0;
let preventPullToRefresh = false;

$('body').on('touchstart', function (e) {
    if (e.originalEvent.touches.length != 1)
        return;

    lastTouchY = e.originalEvent.touches[0].clientY;
    preventPullToRefresh = window.pageYOffset == 0;
});

$('body').on('touchmove', function (e) {
    const touchY = e.originalEvent.touches[0].clientY;
    const touchYDelta = touchY - lastTouchY;

    lastTouchY = touchY;
    if (preventPullToRefresh) {
        preventPullToRefresh = false;
        if (touchYDelta > 0) {
            e.preventDefault();
            return;
        }
    }
});

////////////////////////////////////////////////////////////////////////////////////////////////////////
// passt die Header Tab Anzeige beim scrollen an
////////////////////////////////////////////////////////////////////////////////////////////////////////
headerVisibility('true');
function headerVisibility(_state) {
    let lastScrollTop = 0;

    $('.content').scroll(function(_event){
        var st = $(this).scrollTop();

        if (st > lastScrollTop) { // downscroll code
            $('.content_single #tabs').hide();
            $('.content_big #tabs').hide();
        }
        else { // upscroll code
            $('.content_single #tabs').show();
            $('.content_big #tabs').show();
        }
        lastScrollTop = st;
    });
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// triggert beim Logoff eines Users das Löschen des Cookies
////////////////////////////////////////////////////////////////////////////////////////////////////////
function user_logoff() {
    window.location.href = 'signin.php?logout=1';
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function i18n(key) {
    for (let i of [interfaceData.i18n, interfaceData.i18n.default])
        if (i !== undefined && i[key] !== undefined)
            return i[key];

    return '?';
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function condition_check(cond, values) {
    if (cond === undefined || cond === null)
        return true;

    if (cond.not !== undefined)
        return ! condition_check(cond.not, values);

    if (cond.and !== undefined)
        return cond.and.map(x => condition_check(x, values))
                       .every(x => x === true);

    if (cond.or !== undefined) {
        for (const i of cond.or)
            if (condition_check(i, values) === true)
                return true;
        return false;
    }

    if (cond.operator === undefined)
        return false;

    const value = typeof values === 'object' && cond.index !== undefined
                    ? values[cond.index].value
                    : values;

    switch (cond.operator) {
        case 'not':
        case 'ne':      return value != cond.value;
        case 'e':
        case 'eq':      return value == cond.value;
        case 'g':
        case 'gt':      return value  > cond.value;
        case 'l':
        case 'lt':      return value  < cond.value;
        case 'ge':      return value >= cond.value;
        case 'le':      return value <= cond.value;
        case 't':
        case 'true':    return true;
        case 'f':
        case 'false':
        default:        return false;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function condition_get_matching(rendering, props) {
    for (var i = 0; i < rendering.length; ++i) {
        var cur = rendering[i];

        if (condition_check(cur.condition, props.value))
            return cur.definitions;
    }

    // return empty default
    return {};
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function get_or_default(obj, key, def) {
    return (typeof(obj) == 'object' && key in obj) ? obj[key] : def;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function set_or_extend(arr, idx, vals) {
    if (idx >= arr.length)
        arr.push(vals);
    else {
        var keys = Object.keys(vals);
        for (var i = 0; i < keys.length; ++i)
            arr[idx][keys[i]] = vals[keys[i]];
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// VUE default components
////////////////////////////////////////////////////////////////////////////////////////////////////////
// Clone {{{
// src: https://stackoverflow.com/a/728694
function clone(obj) {
    let copy;

    // Handle the 3 simple types, and null or undefined
    if (null == obj || 'object' != typeof obj)
        return obj;

    // Handle Date
    if (obj instanceof Date) {
        copy = new Date();
        copy.setTime(obj.getTime());
        return copy;
    }

    // Handle Array
    if (obj instanceof Array) {
        copy = [];
        for (let i = 0, len = obj.length; i < len; i++)
            copy[i] = clone(obj[i]);

        return copy;
    }

    // Handle Object
    if (obj instanceof Object) {
        copy = {};
        for (let attr in obj) {
            if (obj.hasOwnProperty(attr)) copy[attr] = clone(obj[attr]);
        }
        return copy;
    }

    throw new Error('Unable to copy obj! Its type isn\'t supported.');
}
// }}}



// Mixins {{{
const mixin_device_control_wrapper = {
    inject: {
        layer: {default: 3},
    },

    computed: {
        device_control_wrapper: function () {
            return this.layer > 2 ? 'control_wrapper' : 'device_wrapper';
        },
    },
};
// }}}



// Utils {{{
Vue.component('TODO', {
    abstract: true,

    template: `<span style="color: red">TODO: <slot /></span>`
});



Vue.component('shif-title', {
    props: {
        classname: {type: String, default: undefined},
        disabled: {
            type: Object,
            default: () => ({flag: false})
        },
    },
    template: `
        <div class="device_title" v-bind:class="classname">
            <slot></slot>
            <span v-if="disabled.flag" class="disabled_text">
                {{ disabled.texts.title }}
            </span>
        </div>
    `,
});



Vue.component('shif-status', {
    props: [
        'classname',
        'key_vals'
    ],
    template: `
        <div class="device_status" v-bind:class="classname">
            <template v-if="key_vals">
                <template v-if="key_vals.length === 0">
                    <div class="device_status_entry"></div>
                </template>
                <template v-else>
                    <span v-for="i in key_vals" class="device_status_entry">
                        <span v-if="i.key" class="name">{{ i.key }}:</span>

                        <template v-if="typeof(i.value) === 'object' && i.value.type === 'color'">
                            <span class="status_color_bullet"
                                  v-bind:style="{backgroundColor: i.value.color}">
                            </span>
                        </template>
                        <template v-else>
                            <span class="value">{{ i.value }}</span>
                        </template>
                    </span>
                </template>
            </template>
            <template v-else>
                <div class="device_status_entry">
                    <slot></slot>
                </div>
            </template>
        </div>
    `,
});



Vue.component('shif-icon', {
    props: {
        src: {
            type: String,
            required: true,
        },
        classname: String,
        active:    String,
        rotate:    Boolean,
    },

    template: `
        <div v-bind:class="classname" v-on:click="$emit('click', this)">
            <template v-if="src in interfaceIcons">
                <div class="svg_icon"
                     v-bind:class="[src, active, {accordion_arrow_rotated: rotate}]"
                     v-html="interfaceIcons[src]">
                </div>
            </template>
            <template v-else-if="src in interfaceData.iconFallback">
                <div class="svg_icon"
                     v-bind:class="[src, active, {accordion_arrow_rotated: rotate}]"
                     v-html="interfaceIcons[interfaceData.iconFallback[src]]">
                </div>
            </template>
            <template v-else>
                <div class="svg_icon icon_default">
                    <svg xmlns="http://www.w3.org/2000/svg" version="1.1" x="0" y="0" width="500" height="500" viewBox="0 0 500 500">
                        <path d="M227.52 320.99L227.52 301.54Q227.52 289.25 229.31 279.52 231.1 269.79 234.94 261.09 238.78 252.38 245.18 244.19 251.58 236 261.31 226.78L284.35 204.77Q294.08 196.06 300.99 185.31 307.9 174.56 307.9 159.2 307.9 139.23 295.36 125.66 282.82 112.1 259.78 112.1 248 112.1 238.53 116.45 229.06 120.8 222.14 128.22 215.23 135.65 211.65 145.38 208.06 155.1 207.55 165.34L144.58 159.71Q147.65 135.14 157.63 115.94 167.62 96.74 183.49 83.42 199.36 70.11 219.84 63.2 240.32 56.29 263.87 56.29 285.89 56.29 305.6 62.69 325.31 69.09 340.42 81.63 355.52 94.18 364.22 112.61 372.93 131.04 372.93 155.1 372.93 171.49 369.34 183.78 365.76 196.06 359.1 206.3 352.45 216.54 342.98 226.02 333.5 235.49 321.73 245.73 311.49 254.43 305.09 261.09 298.69 267.74 294.85 274.4 291.01 281.06 289.47 288.74 287.94 296.42 287.94 307.68L287.94 320.99zM217.28 392.16Q217.28 375.78 229.31 364 241.34 352.22 258.24 352.22 274.62 352.22 286.91 363.49 299.2 374.75 299.2 391.14 299.2 407.52 287.17 419.3 275.14 431.07 258.24 431.07 250.05 431.07 242.62 428 235.2 424.93 229.57 419.81 223.94 414.69 220.61 407.52 217.28 400.35 217.28 392.16z"/>
                    </svg>
                </div>
            </template>
        </div>
    `,
});



Vue.component('shif-slider', {
    mixins: [mixin_device_control_wrapper],

    props: {
        min:   Number,
        max:   Number,
        unit:  String,
        value: Number,
        title: String,
        step:  Number,
        disabled: {
            type: Object,
            default: () => ({flag: false})
        },
        precision: {
            type: Number,
            default: 0,
        }
    },

    computed: {
        value_formatted: function () {
            return this.float_formatted(this.value, this.precision);
        },
    },

    template: `
        <div v-bind:class="[device_control_wrapper, {disabled: disabled.flag}]">
            <div class="device slider">
                <div class="checkbox_wrapper">
                    <div v-if="$slots.profiles"
                        class="checkbox_profiles_wrapper">
                        <slot name="profiles" />
                    </div>
                    <div v-if="$slots.automations" class="checkbox_automation_wrapper">
                        <slot name="automations" />
                    </div>
                    <div v-if="$slots.draggable" class="checkbox_draggable_wrapper">
                        <slot name="draggable" />
                    </div>
                </div>
                <shif-title v-bind:disabled="disabled">{{ title }}</shif-title>
                <div class="slider_action">
                    <div class="amount">
                        <p>{{ value_formatted }} {{ unit }}</p>
                    </div>
                </div>

                <input type="range"
                        class="range"
                        name="range"
                        v-bind:step="step"
                        v-bind:min="min"
                        v-bind:max="max"
                        v-bind:value="value"
                        v-bind:disabled="disabled.flag"
                        v-on:change="$emit('change', parseFloat($event.target.value))"
                        v-on:input="$emit('input', parseFloat($event.target.value))" />

                <div class="slider_marks">
                    <div class="left">
                        {{ min }} {{ unit }}
                    </div>
                    <div class="right">
                        {{ max }} {{ unit }}
                    </div>
                </div>
            </div>
        </div>
    `,
});



Vue.component('shif-radio', {
    mixins: [mixin_device_control_wrapper],

    props: {
        title:     String,
        classname: String,
        values:    Array,
        disabled: {
            type: Object,
            default: () => ({flag: false})
        },
    },

    computed: {
        identifier: function () {
            return 'x' + Math.random().toString(36).substring(10);
        },
    },

    template: `
        <div v-bind:class="[device_control_wrapper, {disabled: disabled.flag}]">
            <div class="device">
                <div class="checkbox_wrapper">
                    <div v-if="$slots.profiles"
                        class="checkbox_profiles_wrapper">
                        <slot name="profiles" />
                    </div>

                    <div v-if="$slots.automations" class="checkbox_automation_wrapper">
                        <slot name="automations" />
                    </div>

                    <div v-if="$slots.draggable" class="checkbox_draggable_wrapper">
                        <slot name="draggable" />
                    </div>
                </div>

                <shif-title v-bind:disabled="disabled">{{ title }}</shif-title>
                <div class="device_radio">
                    <template v-for="i in values">
                        <label class="rad">
                            <span class="rad_name">{{ i.name }}</span>
                            <input type="radio"
                                   v-bind:name="identifier"
                                   v-bind:value="i.value"
                                   v-bind:checked="i.selected"
                                   v-bind:disabled="disabled.flag"
                                   v-on:input="$emit('input', $event.target.value)" />
                            <i></i>
                        </label>
                    </template>
                </div>
            </div>
        </div>
    `,
});



Vue.component('shif-dropdown', {
    mixins: [mixin_device_control_wrapper],

    props: {
        title:     String,
        classname: String,
        values:    Array,
        selected:  [String, Number],
        disabled: {
            type: Object,
            default: () => ({flag: false})
        },
    },

    data: function () {
        return {
            selected_entry: this.selected,
        };
    },

    template: `
        <div v-bind:class="[device_control_wrapper, {disabled: disabled.flag}]">
            <div class="device">
                <div class="checkbox_wrapper">
                    <div v-if="$slots.profiles"
                        class="checkbox_profiles_wrapper">
                        <slot name="profiles" />
                    </div>
                    <div v-if="$slots.automations" class="checkbox_automation_wrapper">
                        <slot name="automations" />
                    </div>
                    <div v-if="$slots.draggable" class="checkbox_draggable_wrapper">
                        <slot name="draggable" />
                    </div>
                </div>
                <shif-title v-bind:disabled="disabled">{{ title }}</shif-title>
                <div class="device_dropdown">
                    <select v-bind:class="{disabled: disabled.flag}"
                            v-model="selected_entry"
                            v-on:change="$emit('change', selected_entry)">
                        <option v-for="i in values"
                                v-bind:value="i.value">
                            {{ i.name }}
                        </option>
                    </select>
                </div>
            </div>
        </div>
    `,
});



Vue.component('shif-button', {
    props: {
        width: {
            type: String,
            default: '100%',
        },
        classname: String,
        disabled: {
            type: Object,
            default: () => ({flag: false})
        },
    },
    template: `
        <div class="control_button"
             v-bind:class="{[classname]: true, disabled: disabled.flag}"
             v-bind:style="{width}"
             v-on:click="(!disabled.flag) && $emit('click', 1)">
            <slot />
            <div class="checkbox_wrapper">
                <div v-if="$slots.draggable" class="checkbox_draggable_wrapper">
                    <slot name="draggable" />
                </div>
                <div v-if="$slots.automations" class="checkbox_automation_wrapper">
                    <slot name="automations" />
                </div>
            </div>
        </div>
    `,
});



Vue.component('shif-colorpicker', {
    mixins: [mixin_device_control_wrapper],

    props: {
        width:  { type: [Number, Object], required: true, },
        height: { type: [Number, Object], required: true, },
        color:  { type: String, required: true, },
        padding:      { type: Number, default:  1 },
        borderWidth:  { type: Number, default:  3 },
        handleRadius: { type: Number, default: 12 },
        sliderMargin: { type: Number, default: 24 },
        sliderHeight: { type: Number, default: 32 },
        borderColor:  { type: String, default: '#fff' },
        anticlockwise:  { type: Boolean, default: true },
        title:  { type: String, default: ''},
        disabled: {
            type: Object,
            default: () => ({flag: false})
        },
    },

    data: function () {
        return {
            handle: null,
        };
    },

    watch: {
        color: function (color_new) {
            if (this.handle)
                this.handle.color.hexString = color_new;
        },
    },

    mounted: function () {
        function rel_to_abs_px(dim_elem, dim_wanted) {
            let result = dim_elem;

            if (typeof(dim_wanted) === 'number')
                return result * (dim_wanted / 100);

            if (typeof(dim_wanted) === 'object' &&
                'percent' in dim_wanted &&
                typeof(dim_wanted.percent) === 'number') {

                result *= (dim_wanted.percent / 100);

                if ('max_pixels' in dim_wanted &&
                    typeof(dim_wanted.max_pixels) === 'number')
                    result = Math.min(result, dim_wanted.max_pixels);
            }

            return result - 30;
        }

        const elem = this.$refs.colorpicker;
        this.handle = new iro.ColorPicker(elem, {
            width:         rel_to_abs_px(elem.scrollWidth,  this.width),
            height:        rel_to_abs_px(elem.scrollHeight, this.height),
            color:         this.color,
            title:         this.title,
            handleRadius:  this.handleRadius,
            padding:       this.padding,
            sliderMargin:  this.sliderMargin,
            sliderHeight:  this.sliderHeight,
            borderWidth:   this.borderWidth,
            borderColor:   this.borderColor,
            anticlockwise: this.anticlockwise,
            display:       'block',
            layout:        [
                {
                    component: iro.ui.Wheel,
                    options: {}
                },
                {
                    // regular value slider
                    component: iro.ui.Slider,
                    options: {}
                },
                {
                    // hue slider
                    component: iro.ui.Slider,
                    options: {
                        sliderType: 'hue'
                    }
                },
                {
                    // saturation slider
                    component: iro.ui.Slider,
                    options: {
                        sliderType: 'saturation'
                    }
                }
            ]
        });

        // `on` patches `this`.
        // So we need to back it up first.
        let outer = this;
        this.handle.on('color:change', function (color, changes) {
            outer.$emit('color:change', {color: color, changes: changes});
        });
        this.handle.on('input:start', function (color) {
            outer.$emit('input:start', {color: color});
        });
        this.handle.on('input:end', function (color) {
            outer.$emit('input', color.hexString);
        });
    },

    template: `
        <div v-bind:class="[device_control_wrapper, {disabled: disabled.flag}]">
            <div class="device color">
                <div class="checkbox_wrapper">
                    <div v-if="$slots.profiles"
                        class="checkbox_profiles_wrapper">
                        <slot name="profiles" />
                    </div>
                    <div v-if="$slots.automations" class="checkbox_automation_wrapper">
                        <slot name="automations" />
                    </div>
                    <div v-if="$slots.draggable" class="checkbox_draggable_wrapper">
                        <slot name="draggable" />
                    </div>
                </div>
                <shif-title v-if="title">{{ title }}</shif-title>
                <div ref="colorpicker">
                </div>
            </div>
        </div>
    `
});



Vue.component('shif-checkbox', {
    props: {
        value: {
            type: Boolean,
            required: true
        },
    },
    template: `
        <label class="check">
            <input type="checkbox"
                   v-bind:checked="value"
                   v-on:click="$emit('click', $event.target.checked)"
                   v-on:input="$emit('input', $event.target.checked)">
            <span class="checkmark"></span>
        </label>
    `,
});
// }}}



// Generic l2 {{{
Vue.component('shif-generic-l2', {
    mixins: [mixin_device_control_wrapper],

    props: {
        icon:        String,
        place:       String,
        title:       String,
        status:      [Array, String],
        actions:     Boolean,
        icon_rotate: Boolean,
        accordion:   Boolean,
        disabled: {
            type: Object,
            default: () => ({flag: false})
        },
        active:  {
            type: Object,

            default: function () {
                return {
                    icon: '',
                    text: '',
                };
            },
        },
    },

    data: function () {
        return {
            click_icon_patched: this.$listeners &&
                                this.$listeners.click !== undefined &&
                                this.$listeners.click_icon === undefined,
        };
    },

    mounted: function () {
        if (this.click_icon_patched)
            this.$on('click_icon', this.$listeners.click);
    },

    beforeDestroy: function () {
        if (this.click_icon_patched)
            this.$off('click_icon', this.$listeners.click);
    },

    methods: {
        emit: function (key, val) {
            if (this.disabled.flag)
                return;

            if (val === undefined)
                this.$emit(key);
            else
                this.$emit(key, val);
        },
    },

    template: `
        <div v-bind:class="[device_control_wrapper, {disabled: disabled.flag}]"
             v-on:mousedown="emit('mousedown')"
             v-on:mouseup="emit('mouseup')"
             v-on:click="emit('click')">
            <div class="device">
                <div class="checkbox_wrapper">

                    <div v-if="$slots.draggable" class="checkbox_draggable_wrapper">
                        <slot name="draggable" />
                    </div>

                    <div v-if="$slots.favorites" class="checkbox_favorites_wrapper">
                        <slot name="favorites" />
                    </div>

                    <div v-if="$slots.profiles" class="checkbox_profiles_wrapper">
                        <slot name="profiles" />
                    </div>

                    <div v-if="$slots.automations" class="checkbox_automation_wrapper">
                        <slot name="automations" />
                    </div>

                </div>

                <div v-on:click.stop="emit('click_icon')">
                    <shif-icon v-bind:src="icon"
                               v-bind:active="active.icon"
                               classname="device_icon">
                    </shif-icon>
                </div>
                <div class="device_text">
                    <shif-title v-bind:disabled="disabled">{{ title }}</shif-title>
                    <div v-if="place" class="device_location">
                        {{ place }}
                    </div>
                    <template v-if="typeof(status) === 'object'">
                        <shif-status v-bind:classname="active.text"
                                     v-bind:key_vals="status">
                        </shif-status>
                    </template>
                    <template v-else>
                        <shif-status v-bind:classname="active.text">{{ status }}</shif-status>
                    </template>
                </div>
                <div v-if="actions"
                     class="device_action">
                    <template v-if="accordion">
                        <shif-icon src="arrow_down_1" v-bind:rotate="icon_rotate"></shif-icon>
                    </template>
                    <template v-else>
                        <div class="svg_icon arrow_right_1">
                            <svg xmlns="http://www.w3.org/2000/svg" version="1.1" id="svg" x="0" y="0" width="370.81" height="370.81" viewBox="0 0 370.81 370.81">
                                <g id="Ebene_1">
                                    <path d="M77.9 345.97L102.03 370.81 292.92 185.41 102.03 0 77.9 24.85 243.18 185.41z"/>
                                </g>
                            </svg>
                        </div>
                    </template>
                </div>
            </div>
        </div>
    `,
});
// }}}



// [shif_device] Generic Shif Device Component Object {{{
function status_impl(control, layer) {
    let out = [];

    const key = (control.texts &&
                 control.texts.l2_state_title &&
                 control.texts.l2_state_title.content) ?
        control.texts.l2_state_title.content :
        null;

    if (control.metadata !== undefined &&
        control.metadata.statusColor === true &&
        control.variableInputs.length > 0)
        return [{
            key: key,
            value: {
                type: 'color',
                color: control.variableInputs[0].properties.value,
            },
        }];

    let val = [];
    for (const input of control.variableInputs) {
        if (input.properties.visualizeInOverview === false && layer == 2)
            continue;

        if (input.rendering) {
            const sel = condition_get_matching(input.rendering, input.properties);
            if (sel.texts !== undefined &&
                sel.texts.state !== undefined &&
                sel.texts.state.content !== undefined) {
                val.push(sel.texts.state.content);
                continue;
            }
        }

        const unit = input.properties.unit ? input.properties.unit : '';
        val.push(input.properties.value + ' ' + unit);
    }

    if (val.length > 0)
        out.push({key: key, value: val.join(', ')});

    return out;
}



const shif_device = {
    props: [
        'uiElement',
        'control',
        'device',
        'icons',
        'texts',
        'output',
        'props',
        'indexes',
        'rendering',
        'include_place',
        'sibling_idx',
    ],

    mixins: [mixin_modemenu],

    data: function() {
        return {
            lastClickCount: 0,
            profile_state: false,
        };
    },

    inject: {
        layer: 'layer',
        role_id: {default: undefined,},
        room_id: 'room_id',
        floor_id: 'floor_id',
        // siblings: 'siblings',
    },

    methods: {
        status_minimal: function (descs=true) {
            const raw = status_impl(this.control, this.layer);

            return (descs) ? raw : raw.map(x => ({value: x.value}));
        },
    },

    computed: {
        cond: function () {
            const sel = condition_get_matching(this.rendering, this.props);

            return {
                text: get_or_default(sel.texts, 'state', {content: '', color: ''}),
                icon: get_or_default(sel.icons, 'state', {name:    '', color: ''}),
            };
        },

        dev: function () {
            return interfaceData.devices[this.device];
        },

        place: function () {
            if (!this.include_place)
                return null;

            const room = interfaceData.rooms[this.dev.room];
            if (interfaceData.options.showFloor !== true)
                return room.name;

            if (room.floors.length === 0)
                return `${i18n('house.storyless')} - ${room.name}`;

            return room.floors.map(x => interfaceData.floors[x])
                              .map(x => x.name + ' - ' + room.name)
                              .join(' | ');
        },

        title: function () {
            if (this.control.texts &&
                this.control.texts.title &&
                this.control.texts.title.content)
                return this.control.texts.title.content;

            return this.dev.label;
        },

        status: function () {
            let out = [];

            for (const control of this.dev.controls)
                out = out.concat(status_impl(control, this.layer));

            return out;
        },

        disabled: function () {
            const backend = check_disabled_backend(this.uiElement, this.indexes);
            if (backend.flag === true)
                return backend;

            return check_disabled_frontend(this.uiElement, this.sibling_idx,
                                           this.$parent.dev_obj_props);
        },

        used_by_automations: function () {
            if (interfaceData.map_automation === undefined ||
                interfaceData.map_automation.devices === undefined)
                return false;

            const map = interfaceData.map_automation.devices;
            if (map[this.device] === undefined)
                return false;

            const dev = map[this.device];
            if (dev[this.indexes.control] === undefined)
                return false;

            const control = dev[this.indexes.control];
            if (control[this.indexes.input] === undefined)
                return false;

            const input = control[this.indexes.input];
            if(input.lenght === 0)
                return false;

            return input;
        },

        automation_link: function () {
            if (this.used_by_automations === false)
                return {};

            return this.used_by_automations.length === 1
                    ? {
                        name: 'settings.automations.automation',
                        params: {
                            automation_id: this.used_by_automations,
                        },
                      }
                    : {
                        name: 'settings.automations.selection',
                        params: {
                            automation_ids: this.used_by_automations.join('-'),
                        },
                      };
        }
    },
};



function shif_comps_create(name, l2, l3) {
    const shif_name = 'shif-' + name + '-';

    Vue.component(shif_name + 'l2', l2);
    Vue.component(shif_name + 'l3', l3);
}



function shif_register_disable_hooks(objs) {
    if (interfaceData.disable_hooks === undefined)
        Vue.set(interfaceData, 'disable_hooks', {});

    for (const i in objs)
        Vue.set(interfaceData.disable_hooks, i, objs[i]);
}
// }}}



Vue.component('shif-room', {
    props: {
        floor: Object,
        room:  [String, Number],
    },

    mixins: [mixin_modemenu],

    methods: {
        link: function (floor_key, room_val) {
            return {
                name: 'house.tab.rooms.room',
                params: {
                    floor_id: floor_key,
                    room_id:  room_val,
                },
            };
        },
    },

    template: `
        <div class="roomSelect_wrapper">
            <router-link v-bind:to="link(floor.key, room)">
                <shif-icon v-bind:src="interfaceData.rooms[room].icon"
                           class="roomSelect" />
                <shif-icon v-if="modemenu_is_state('DRAGGABLE')"
                           src="move_2" classname="drag_drop_icon_right" />
                <div class="description">
                    {{ interfaceData.rooms[room].name }}
                </div>
            </router-link>
        </div>
    `,
});



Vue.component('shif-tab', {
    props: {
        width: {
            type:    String,
            default: '50%',
        }
    },
    template: `
        <div class="tab button"
             v-bind:style="{width: width}"
             v-on:click="$emit('click', 1)">
             <slot></slot>
        </div>
    `,
});

/*
    global
        ModeMenuState
        get_or_default
        mixin_modemenu
        scroll_positions
*/
/*
    exported
        mixin_menus
        mixin_rooms
        mixin_scroll_position
*/



function comp_obj(control, device, input, output, is, indexes) {
    let ret = {
        uiElement: device,
        control:   control,
        device:    device.databaseId,
        icons:     device.icons,
        texts:     device.texts,
        output:    output,
        props:     input.properties,
        indexes:   indexes,
        rendering: input.rendering,
    };

    if (is)
        ret.is = is;

    return ret;
}



const mixin_rooms = {
    computed: {
        unassigned_rooms: function () {
            const tmp = Object.keys(interfaceData.floors)
                              .map(x => interfaceData.floors[x].rooms)
                              .flat();

            let assigned = {};
            for (const i of tmp)
                assigned[i] = true;

            return Object.keys(interfaceData.rooms)
                         .filter(x => ! assigned[x])
                         .filter(x => this.room_has_devices(x));
        },
    },

    methods: {
        room_has_devices: function (room_key) {
            const room = interfaceData.rooms[room_key];

            return room.devices !== undefined && room.devices.length > 0;
        },
    }
};


const mixin_menus = {
    methods: {
        disabled: function (type, key) {
            const key_type = ({
                'mainmenu': 'disabledMainmenus',
                'menu':     'disabledMenus',
            })[type];

            return interfaceData.options &&
                   interfaceData.options[key_type] &&
                   interfaceData.options[key_type][key] === true;
        },
    },
};



const mixin_unique_view_key = {
    methods: {
        _specific_bc_idx_wanted: function (mode) {
            const ident = this.$route.meta.cache_ident;

            return ident !== undefined &&
                   ident[mode] !== undefined &&
                   ident[mode].bc_idx !== undefined;
        },

        _uniqueify_needed: function (mode) {
            const ident = this.$route.meta.cache_ident;

            return ident !== undefined &&
                   ident[mode] !== undefined &&
                   ident[mode].params !== undefined &&
                   ident[mode].params.length > 0;
        },

        _base: function (mode) {
            const meta = this.$route.meta;
            const breadcrumbs = meta.breadcrumbs;

            if (breadcrumbs === undefined)
                return this.$route.fullPath;

            if (this._specific_bc_idx_wanted(mode)) {
                const idx = meta.cache_ident[mode].bc_idx;

                if (breadcrumbs[idx] !== undefined)
                    return breadcrumbs[idx];
            }

            switch (mode) {
                case 'default': // fall through
                case 'single': // fall through
                case 'big':    return breadcrumbs[breadcrumbs.length - 1];
                case 'small':  return breadcrumbs[breadcrumbs.length - 2];
            }

            throw `BUG: mixin_unique_view_key: mode is unknown: ${mode}`;
        },

            // Make base unique over the rooms/devices/etc.
        _uniqueify: function (old, mode) {
            const cache_ident = this.$route.meta.cache_ident;

            if (! this._uniqueify_needed(mode))
                return old;

            const unique = cache_ident[mode].params
                                            .map(x => this.$route.params[x])
                                            .join('-');

            return `${old}-${unique}`;
        },

        /**
         * It's hard to defined a unique key for each view part.
         * Requirements:
         *   * lvl(1).default/single == lvl(2).small
         *   * lvl(n).big == lvl(n + 1).small
         **/
        unique_view_key: function (mode) {
            /**
             * To prevent me from debugging this function another time...
             * The `__transition-\d+-` part in `$vnode.key` is added somewhere
             * else!
             **/
            const base = this._base(mode);
            return this._uniqueify(base, mode);
        },

        find_unique_view_key: function () {
            let cur = this;

            while(cur.$vnode.key === undefined && cur.$parent !== undefined)
                cur = cur.$parent;

            const key = cur.$vnode.key;
            if (key === undefined)
                return key;

            return key.replace(/^__transition-\d+-/, '');
        }
    },
};



const mixin_scroll_position = {
    methods: {
        // _find_scroll_position_marker: function (start) {
        //     let cur = start;
        //
        //     for (; cur !== undefined; cur = cur.$parent) {
        //         if (cur.$options.name === 'shif-scroll-position-marker')
        //             return cur;
        //     }
        //
        //     return cur;
        // },

        _key: function (vm) {
            return vm.$vnode.key.replace(/^__transition-\d+-/, '');
        },
    },

    beforeRouteLeave: function (to, from, next) {
        const key = this._key(this);

        scroll_positions[key] = {
            x: this.$el.scrollLeft,
            y: this.$el.scrollTop,
        };

        if (key === 'house.tab.devices')
            scroll_positions[key].role_id = this.$refs.devices.role_id_opened;

        next();
    },

    beforeRouteEnter: function (to, from, next) {
        next(vm => {
            const key = vm._key(vm);

            if (scroll_positions[key] === undefined)
                return;

            const pos = scroll_positions[key];

            vm.$el.scroll(pos.x, pos.y);
        });
    },
};



const mixin_components = {
    methods: {
        find_component: function (device, layer) {
            if (layer == 'l2' && typeof(device.metadata) == 'object') {
                if ('l2_action' in device.metadata) {
                    const keys = device.metadata.l2_action;

                    const control = device.controls[keys.control];
                    const input   = control.variableInputs[keys.input];
                    const output  = control.variableOutputs[keys.input];
                    const is      = 'shif-' + control.control + '-l2';

                    return [comp_obj(control, device, input, output, is,
                                     {input: keys.input, control: keys.control})];
                }

                if (device.controls.length <= 1 &&
                    (!('l3_force' in device.metadata) ||
                     device.metadata.l3_force !== true))
                    layer = 'l3';
            }

            let out = [];
            // for (const control of device.controls) {
            for (let i = 0; i < device.controls.length; ++i) {
                const control = device.controls[i];

                for (const k in control.variableInputs) {
                    const input  = control.variableInputs[k];
                    const output = control.variableOutputs[k];
                    const is     = 'shif-' + control.control + '-' + layer;

                    out.push(comp_obj(control, device, input, output, is,
                                      {input: k, control: i}));
                }
            }

            return out;
        },
    },
};



const mixin_print_mounted = function () {
    return {
        // This is for debug reasons only
        // mounted: function () {
            // console.log(this.$options.name || arguments[0]);
        // },
    };
};



function root_profiles_idx(peer, channel, name) {
    return `${peer}.${channel}.${name}`;
}


const mixin_profiles = {
    computed: {
        favorite_profiles: function () {
            return Object.keys(interfaceData.profiles)
                         .map(x => interfaceData.profiles[x])
                         .filter(x => x.favorite === true);
        },

        global_profiles: function () {
            return Object.keys(interfaceData.profiles)
                         .map(x => interfaceData.profiles[x])
                         .filter(x => x.global === true);
        },

        local_profiles: function () {
            const floor = parseInt(this.floor_id);
            const room  = parseInt(this.room_id);

            return Object.keys(interfaceData.profiles)
                         .map(x => interfaceData.profiles[x])
                         .filter(
                x => x.locations.findIndex(
                    loc => {
                        return loc.floorId === floor &&
                          (loc.roomId === undefined || loc.roomId === room);
                    }
                ) !== -1
            );
        },

        editable_profiles: function () {
            return Object.keys(interfaceData.profiles)
                         .map(x => interfaceData.profiles[x])
                         .filter(x => x.editable !== false);
        },
    },

    methods: {
        locations: function (floor, room) {
            if (floor === undefined && room === undefined)
                return [];

            return [
                {
                    floorId: floor === null ? undefined : Number(floor),
                    roomId:  room  === null ? undefined : Number(room),
                }
            ];
        },

        role_profiles: function (role_id) {
            return Object.keys(interfaceData.profiles)
                         .map(x => interfaceData.profiles[x])
                         .filter(x => x.roles !== undefined &&
                                      x.roles.length > 0 &&
                                      x.roles.findIndex(x => x.role === Number(role_id)) !== -1);
        },

        profile_build_root_devs: function (profile) {
            // Overwrite current profile on new load
            this.$root.profiles.devs = {};

            for (const i of profile.values) {
                const idx = root_profiles_idx(i.peerId, i.channel, i.variable);
                this.$root.profiles.devs[idx] = {
                    peer:    i.peerId,
                    channel: i.channel,
                    name:    i.variable,
                    value:   i.value,
                };
            }
        },

        profile_icon: function (profile) {
            return get_or_default(profile, 'icon', 'slider_1');
        },

        profile_load: function (profile, cb) {
            this.profile_start(profile, (result) => {
                this.modemenu_show(ModeMenuState.PROFILES);

                if (cb)
                    return cb(result);
            });
        },

        profile_start: function (profile, cb) {
            return this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'activateVariableProfile',
                params: [profile.id],
            }, cb);
        },

        profile_delete: function (profile, cb) {
            return this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'deleteVariableProfile',
                params: [profile.id],
            }, (result) => {
                Vue.delete(interfaceData.profiles, this.profile.id);

                if (this.modemenu_is_state(ModeMenuState.PROFILES))
                    this.modemenu_hide();

                if (cb)
                    return cb(result);
            });
        },

        profile_add: function (form, cb) {
            const locations = this.locations(form.location.floor,
                                             form.location.room);

            return this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'addVariableProfile',
                params: [
                    {
                        [interfaceData.options.language]: form.profile_name,
                    },
                    {
                        global:    form.location.global,
                        favorite:  form.location.favorite,
                        icon:      form.icon,
                        locations: locations,
                        roles:     [],
                        values:    [],
                    }
                ],
            }, (result) => {
                Vue.set(interfaceData.profiles, result.result, {
                    id:        result.result,
                    icon:      form.icon,
                    locations: locations,
                    global:    form.location.global,
                    favorite:  form.location.favorite,
                    name:      form.profile_name,
                    roles:     [],
                    values:    [],
                });

                this.modemenu_hide();

                if (cb)
                    return cb(result);
            });
        },

        profile_update: function (profile, form, cb) {
            function val_into_proper_type(value) {
                if (value === 'true')
                    return true;
                if (value === 'false')
                    return false;

                const as_nr = Number(value);
                if (! Number.isNaN(as_nr))
                    return as_nr;

                return value;
            }

            const locations = this.locations(form.location.floor,
                                             form.location.room);

            const [roles, values] = form.role.role !== null &&
                                    form.role.value !== undefined &&
                                    form.role.value !== null
                ? [[{
                        role: Number(form.role.role),
                        value: val_into_proper_type(form.role.value)
                  }], []]
                : [[], Object.keys(this.$root.profiles.devs)
                             .map(x => this.$root.profiles.devs[x])
                             .map(x => ({
                                peerId: x.peer,
                                channel: x.channel,
                                variable: x.name,
                                value: x.value,
                             }))];

            return this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'updateVariableProfile',
                params: [
                    profile.id,
                    {
                        [interfaceData.options.language]: form.profile_name,
                    },
                    {
                        global:    form.location.global,
                        favorite:  form.location.favorite,
                        icon:      form.icon,
                        locations: locations,
                        roles:     roles,
                        values:    values,
                    }
                ],
            }, (result) => {
                Vue.set(interfaceData.profiles, profile.id, {
                    id:        profile.id,
                    name:      form.profile_name,
                    global:    form.location.global,
                    favorite:  form.location.favorite,
                    icon:      form.icon,
                    locations: locations,
                    roles:     roles,
                    values:    values,
                });

                this.modemenu_hide();

                if (cb)
                    return cb(result);
            });
        },

        profile_used_by_automations: function (id) {
            return interfaceData.map_automation.profiles[id] === undefined
                    ? false
                    : interfaceData.map_automation.profiles[id];
        },

        profile_automation_link: function (id) {
            const used_by_automations = this.profile_used_by_automations(id);

            if (used_by_automations === false)
                return {};

            return used_by_automations.length === 1
                    ? {
                        name: 'settings.automations.automation',
                        params: {
                            automation_id: used_by_automations,
                        },
                      }
                    : {
                        name: 'settings.automations.selection',
                        params: {
                            automation_ids: used_by_automations.join('-'),
                        },
                      };
        }
    },
};



const mixin_favorites = {
    methods: {
        dev_toggle_favorite: function (dev, state) {
            this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'getUiElementMetadata',
                params: [dev],
            }, (data) => {
                let new_metadata = {};
                if (Array.isArray(data.result))
                    new_metadata = data.result;

                if (new_metadata.favorites === undefined)
                    new_metadata.favorites = {};

                new_metadata.favorites.state = state;

                this.$homegear.invoke({
                    jsonrpc: '2.0',
                    method: 'setUiElementMetadata',
                    params: [dev, new_metadata],
                }, () => {
                    interfaceData.devices[dev].dynamicMetadata.favorites =
                        new_metadata.favorites;
                });
            });
        },
    },
};



Vue.component('shif-trans-drop-down', {
    template: `
        <transition name="trans-drop-down">
            <slot />
        </transition>
    `,
});



Vue.component('shif-trans-right-in-out', {
    template: `
        <transition name="trans-right-in-out" tag="div">
            <slot />
        </transition>
    `,
});



Vue.component('shif-house-collected-entries', {
    mixins: [
        mixin_components,
        mixin_profiles,
        mixin_print_mounted(),
    ],

    props: {
        layer: {
            type: Number,
            required: true,
        },

        favorites: {
            type: Boolean,
            default: false,
        },

        include_place: {
            type: Boolean,
            default: false,
        },
    },

    provide: function () {
        return {
            layer: this.layer,
            siblings: this.dev_objs,
        };
    },

    inject: {
        device_id: { default: undefined, },
        room_id:   { default: undefined, },
        floor_id:  { default: undefined, },
    },

    data: function () {
        return {
            dev_obj_keys: [],
        };
    },

    watch: {
        dev_obj_keys: function () {
            if (this.$root.debug)
                console.log(JSON.stringify(this.dev_obj_keys, null, 4));
        },

        dev_objs: function () {
            if (this.$root.debug)
                console.log(this.dev_objs);
        },
    },

    mounted: function () {
        this.dev_obj_keys = Array.from(this.dev_objs.keys());
    },

    computed: {
        dev_objs: function () {
            if (this.layer === 2) {
                const devices = this.favorites === true
                                    ? Object.keys(interfaceData.devices)
                                            .map(dev => interfaceData.devices[dev])
                                            .filter(dev => dev.dynamicMetadata.favorites &&
                                                           dev.dynamicMetadata.favorites.state)
                                    : interfaceData.rooms[this.room_id]
                                                   .devices
                                                   .map(dev => interfaceData.devices[dev]);

                return devices.map(dev => this.find_component(dev, 'l2'));
            }

            if (this.layer === 3) {
                const device = interfaceData.devices[this.device_id];
                return this.find_component(device, 'l3');
            }

            throw 'Not implemented';
        },

        dev_obj_props: function () {
            return this.dev_objs.map(x => x.props);
        },

        profiles: function () {
            if (this.layer !== 2)
                return [];

            return this.favorites === true
                        ? this.favorite_profiles
                        : this.local_profiles;
        },
    },

    template: `
        <div>
            <div v-if="profiles.length > 0" class="profiles_wrapper">
                <template v-for="i in profiles">
                    <shif-generic-l2 v-bind:icon="profile_icon(i)"
                                     v-bind:title="i.name"
                                     v-bind:status="i18n('modemenu.profiles.name.label')"
                                     v-bind:active="{icon: i.isActive ? 'active' : ''}"
                                     v-on:click="profile_start(i)">
                        <template v-slot:automations>
                            <router-link v-if="profile_used_by_automations(i.id) !== false"
                                         v-bind:to="profile_automation_link(i.id)">
                                <shif-icon src="calendar-time_1" />
                            </router-link>
                        </template>
                    </shif-generic-l2>
                </template>
            </div>

            <shif-draggable v-bind:value="dev_objs"
                            v-slot="draggable"
                            handle=".drag_drop_icon">
                <template v-for="idx in draggable.keys">
                    <component v-if="dev_objs[idx] !== undefined"
                               v-bind="dev_objs[idx]"
                               v-bind:sibling_idx="idx"
                               v-bind:include_place="include_place" />
                </template>
            </shif-draggable>
        </div>
    `,
});



Vue.component('shif-mainmenu-tabs', {
    mixins: [mixin_print_mounted()],

    computed: {
        idx_mainmenu: function () {
            const menu_name = this.$route.matched[0].name;
            return interfaceData.mainmenu.findIndex(x => x.name === menu_name);
        },

        tabs: function () {
            return interfaceData.mainmenu[this.idx_mainmenu].tabs;
        },

        tab_width: function () {
            return (100 / this.tabs.length) + '%';
        },
    },

    template: `
        <div>
            <div id="tabs" v-if="$root.draggable.in_progress === false">
                <template v-for="tab in tabs">
                    <router-link v-bind:to="{name: tab.name}">
                        <shif-tab v-bind:width="tab_width">
                            {{ i18n(tab.name) }}
                        </shif-tab>
                    </router-link>
                </template>
            </div>

            <div class="tabWrapper activeTab" style="text-align: center;">
                <slot />
            </div>
        </div>
    `
});



// This is a helper component, so we have defined parent where we can read the
// scroll positions from.
// Although not defining own DOM elements, it must not be abstract!
// We won't be able to access the scroll positions!
//
// Backup: currently unused.
const ShifScrollPositionMarker = {
    // Explicitly set, so it does not get (accidently?) set to true in the
    // future.
    abstract: false,
    name: 'shif-scroll-position-marker',

    render: function () {
        return this.$slots.default;
    },
};



Vue.component('shif-paging', {
    mixins: [mixin_unique_view_key, mixin_print_mounted()],

    components: {
        ShifScrollPositionMarker,
    },

    computed: {
        is_single_view: function () {
            const idx   = this.$route.matched.length - 1;
            const route = this.$route.matched[idx];

            return route !== undefined &&
                   route.components !== undefined &&
                   route.components.default !== undefined;
        },

        views: function () {
            if (this.is_single_view)
                return [{
                    name:       'default',
                    class:      'content content_single',
                    click:      false,
                    transition: false,
                }];

            return [
                {
                    name:       'small',
                    class:      'content content_small',
                    click:      true,
                    transition: false,
                },
                {
                    name:       'big',
                    class:      'content content_big',
                    click:      false,
                    transition: true,
                }
            ];
        },
    },

    render: function (h) {
        return h('div', {}, this.views.map(i => {
            const router_view = h('router-view', {
                class:    i.class,
                key:      this.unique_view_key(i.name),
                props:    {name: i.name},
                nativeOn: i.click ? {click: () => this.$router.back()} : {}
            });

            if (! i.transition)
                return router_view;

            return h('shif-trans-right-in-out', {}, [router_view]);
        }));
    }
});



Vue.component('shif-checkbox-favorites', {
    mixins: [mixin_favorites, mixin_modemenu],

    props: {
        dev: {
            type: Object,
            required: true,
        },
        classname: {
            type: String,
            default: 'checkbox_favorites',
        },
    },

    inject: ['layer'],

    data: function () {
        return {
            state: this.dev !== undefined &&
                   this.dev.dynamicMetadata !== undefined &&
                   this.dev.dynamicMetadata.favorites !== undefined &&
                   this.dev.dynamicMetadata.favorites.state === true,
        };
    },

    computed: {
        show: function () {
            return this.layer === 2 &&
                   this.modemenu_is_state(ModeMenuState.FAVORITES);
        },
    },

    methods: {
        change: function () {
            return this.dev_toggle_favorite(this.dev.databaseId, this.state);
        },
    },

    template: `
        <div>
            <template v-if="show">
                <div v-on:click.stop=""
                     v-on:change.stop="change">
                    <div v-bind:class="classname">
                        <shif-checkbox v-model="state" />
                    </div>
                </div>
            </template>
        </div>
    `
});



Vue.component('shif-checkbox-profiles', {
    mixins: [mixin_modemenu, mixin_profiles],

    props: {
        dev: {
            type: Object,
            required: true,
        },
        output: {
            type: Object,
            required: true,
        },
        props: {
            type: Object,
            required: true,
        },
        classname: {
            type: String,
            default: '',
        },
    },

    data: function () {
        const idx = root_profiles_idx(this.output.peer, this.output.channel,
                                      this.output.name);

        return {
            idx:   idx,
            state: (idx in this.$root.profiles.devs),
        };
    },

    watch: {
        'props.value': function () {
            if (this.modemenu_is_state(ModeMenuState.PROFILES) &&
                (this.idx in this.$root.profiles.devs))
                this.$root.profiles.devs[this.idx].value = this.props.value;
        },
    },

    methods: {
        change: function () {
            if (this.state)
                this.$root.profiles.devs[this.idx] = {
                    peer:    this.output.peer,
                    channel: this.output.channel,
                    name:    this.output.name,
                    value:   this.props.value,
                };

            else if (this.idx in this.$root.profiles.devs)
                delete this.$root.profiles.devs[this.idx];
        },
    },

    template: `
        <div>
            <template v-if="modemenu_is_state('PROFILES')">
                <div v-on:click.stop=""
                     v-on:change.stop="change">
                    <div v-bind:class="classname">
                        <shif-checkbox v-model="state" />
                    </div>
                </div>
            </template>
        </div>
    `
});



Vue.component('shif-icon-selection', {
    props: {
        value: {type: String, required: true,},
    },

    methods: {
        on_click: function (new_) {
            this.$emit('input', new_);
        },
    },

    template: `
        <div id="profile_icons">
            <label v-for="_, key in interfaceIcons"
                    v-bind:class="{selected: value === key}"
                    v-bind:title="key"
                    class="profile_icon_wrapper">
                <shif-icon classname="profile_icon" v-bind:src="key" />
                <input type="radio"
                        name="profile_icon"
                        v-bind:value="value"
                        v-on:click="on_click(key)"
                        hidden />
            </label>
        </div>
    `
});

/*
    global
        homegear
*/



Vue.component('shif-draggable', {
    // mixins: [mixin_unique_view_key],

    components: {
        draggable: window.vuedraggable,
    },

    props: {
        value: {
            type: [Array, Object],
            default: undefined,
        },
        handle: {
            type: String,
            default: undefined,
        },
        name: {
            type: String,
            default: 'default',
        }
    },

    computed: {
        // Copied from find_unique_view_key
        // Maybe consolidate in future back...
        name_base: function () {
            let cur = this;

            while(cur.$vnode.key === undefined && cur.$parent !== undefined)
                cur = cur.$parent;

            const key = cur.$vnode.key;
            if (key === undefined)
                throw 'BUG: key should be set: ' + key;

            return key.replace(/^__transition-\d+-/, '');
        },

        _interfaceData_field: {
            get: function () {
                if (interfaceData.options.elementOrder === undefined)
                    return undefined;

                const cur = interfaceData.options.elementOrder[this.name_base];
                if (cur === undefined || cur[this.name] === undefined)
                    return undefined;

                return cur[this.name];
            },
            set: function (new_) {
                if (interfaceData.options.elementOrder === undefined)
                    Vue.set(interfaceData.options, 'elementOrder', {});

                if (interfaceData.options.elementOrder[this.name_base] === undefined)
                    Vue.set(interfaceData.options.elementOrder, this.name_base, {});

                Vue.set(interfaceData.options.elementOrder[this.name_base], this.name, new_);
            },
        },

        keys: {
            get: function () {
                return this._interfaceData_field || [];
            },
            set: function (new_) {
                this._interfaceData_field = new_;

                this.$homegear.invoke({
                    jsonrpc: '2.0',
                    method: 'getUserMetadata',
                    params: []
                }, (data) => {
                    let new_settings = data.result;

                    if (new_settings.interface === undefined)
                        new_settings.interface = {};

                    if (new_settings.interface.elementOrder === undefined)
                        new_settings.interface.elementOrder = {};

                    if (new_settings.interface.elementOrder[this.name_base] === undefined)
                        new_settings.interface.elementOrder[this.name_base] = {};

                    new_settings.interface.elementOrder[this.name_base][this.name] = new_;

                    this.$homegear.invoke({
                        jsonrpc: '2.0',
                        method: 'setUserMetadata',
                        params: [new_settings]
                    });
                });
            },
        },

        values: function () {
            return this.keys.map(x => ({
                key: x,
                val: this.value[x],
            }));
        },
    },

    mounted: function () {
        this.init_interfaceData_field();
    },

    watch: {
        'value': function () {
            this.init_interfaceData_field();
        },
    },

    methods: {
        init_interfaceData_field: function () {
            if (typeof this.value !== 'object')
                throw 'Bug: typeof(value) != object: ' + this.value;

            const saved = this._interfaceData_field;

            const new_ = Array.isArray(this.value)
                            ? Array.from(this.value.keys())
                            : Object.keys(this.value);

            this._interfaceData_field =
                saved === undefined || new_.length !== saved.length
                    ? new_
                    : saved;
        }
    },

    template: `
        <draggable v-model="keys"
                   v-bind:handle="handle"
                   v-on:start="$root.draggable.in_progress = true"
                   v-on:end="$root.draggable.in_progress = false">
            <slot v-bind:keys="keys"
                  v-bind:values="values"/>
        </draggable>
    `
});



function _element_order_cleanup() {
    homegear.invoke({
        jsonrpc: '2.0',
        method: 'getUserMetadata',
        params: [],
    }, function (x) {
        let tmp = x.result;
        delete tmp.interface.elementOrder;

        homegear.invoke({
            jsonrpc: '2.0',
            method: 'setUserMetadata',
            params: [tmp],
        });
    });
}

/*
    exported
        NotificationDisplayType
        mixin_notification
*/



const NotificationDisplayType = {
    Error: -1,
    Toast: 0,
    ToastIntegrated: 1,
    Integrated: 2,
    Modal: 3,

    _is_integrated: function (flags) {
        return flags.hasModal === true &&
               flags.overlayModal !== true &&
               flags.overlayNotification !== true;
    },

    _is_modal: function (flags) {
        return flags.hasModal === true &&
               flags.overlayModal === true &&
               flags.overlayNotification !== true;
    },

    _is_toast_only: function (flags) {
        return flags.hasModal !== true &&
               flags.overlayModal !== true &&
               flags.overlayNotification === true;
    },

    _is_toast_integrated: function (flags) {
        return flags.hasModal === true &&
               flags.overlayModal !== true &&
               flags.overlayNotification === true;
    },

    find: function (notification) {
        const flags = typeof(notification) === 'object'
                        ? notification.flags
                        : interfaceData.notifications[notification].flags;

        if (this._is_toast_only(flags))
            return NotificationDisplayType.Toast;

        if (this._is_integrated(flags))
            return NotificationDisplayType.Integrated;

        if (this._is_modal(flags))
            return NotificationDisplayType.Modal;

        if (this._is_toast_integrated(flags))
            return NotificationDisplayType.ToastIntegrated;

        return NotificationDisplayType.Error;
    }
};



function notification_inject_display_type(x) {
    if (x.displayType === undefined)
        Vue.set(x, 'displayType', NotificationDisplayType.find(x));

    return x;
}



const mixin_notification = {
    methods: {
        notification_action: function (id, btn, cb) {
            if (id === undefined)
                id = this.notification_id;

            this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'uiNotificationAction',
                params: [Number(id), Number(btn)],
            }, cb);
        },

        notification_get: function (id, cb) {
            if (id === undefined)
                id = this.notification_id;

            this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'getUiNotification',
                params: [Number(id), interfaceData.options.language],
            }, cb);
        },

        notification_remove: function (id, cb) {
            if (id === undefined)
                id = this.notification_id;

            this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'removeUiNotification',
                params: [Number(id)],
            }, cb);
        },

        notification_hide: function (id, cb) {
            if (id === undefined)
                id = this.notification_id;

            Vue.set(interfaceData.notifications[id], 'seen', true);

            if (cb !== undefined)
                cb();
        },

        notification_is_hidden: function (msg) {
            return msg.seen === true;
        },
    },

    computed: {
        _notifications: function () {
            return Object.keys(interfaceData.notifications)
                         .map(x => interfaceData.notifications[x])
                         .map(notification_inject_display_type);
        },

        modal_notifications: function () {
            return this._notifications
                       .filter(x => x.displayType === NotificationDisplayType.Modal)
                        // filter out messages hidden by buttons[].closeModal
                       .filter(x => ! this.notification_is_hidden(x));
        },

        toast_notifications: function () {
            return this._notifications
                       .filter(
                           x => x.displayType === NotificationDisplayType.Toast ||
                                x.displayType === NotificationDisplayType.ToastIntegrated
                        )
                        // filter out messages hidden by msg.flags.closeable
                       .filter(x => ! this.notification_is_hidden(x));
        },

        integrated_notifications: function () {
            return this._notifications
                       .filter(
                            x => x.displayType === NotificationDisplayType.Integrated ||
                                 x.displayType === NotificationDisplayType.ToastIntegrated
                       );
        },
    },
};

/*
    exported
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
        shif_device_slot_profiles
*/



const shif_device_slot_automations = `
    <template v-slot:automations>
        <router-link v-if="used_by_automations !== false"
                     v-bind:to="automation_link">
            <shif-icon src="calendar-time_1" />
        </router-link>
    </template>
`;



const shif_device_slot_draggable = `
    <template v-slot:draggable>
        <shif-icon v-if="modemenu_is_state('DRAGGABLE')"
                   src="move_1" classname="drag_drop_icon" />
    </template>
`;



const shif_device_slot_favorites = `
    <template v-slot:favorites>
        <shif-checkbox-favorites v-bind:dev="dev" />
    </template>
`;



const shif_device_slot_profiles = `
    <template v-slot:profiles>
        <shif-checkbox-profiles v-bind:dev="dev"
                                v-bind:output="output"
                                v-bind:props="props" />
    </template>
`;

/*
    global
        get_or_default
        homegear
        set_or_extend
*/
/*
    exported
        mixin_states
        states_flag_dirty
*/



let _states = Vue.observable({
    initialized: false,
    content: {},
    dirty: [],
});



function _status_gather_invokes(role, role_id) {
    let ret = [{
        type: get_or_default(role, 'aggregationType', 2),
        ids:  [{'id':role_id,'direction':0}],
    }];

    if (!('rolesInclude' in role))
        return ret;

    for (const role_inc of role.rolesInclude) {
        // TODO: test this refactoring
        // const ids = role_inc.roles.map(x => ({
            // id: x,
            // direction: 0
        // }));
        let ids = [];
        for(const index in role_inc.roles) {
            ids.push({
                id: role_inc.roles[index],
                direction: 0,
            });
        }

        ret.push({
            type: get_or_default(role_inc, 'aggregationType', 2),
            ids:  ids,
        });
    }

    return ret;
}



function _status_text(role, texts_idx) {
    if (role.texts === undefined)
        return '?';

    if ((typeof(texts_idx) === 'number' && role.texts.length > texts_idx) ||
        (typeof(texts_idx) === 'string' && texts_idx in role.texts))
        return role.texts[texts_idx];

    if (role.texts.length == 1)
        return role.texts[0];

    return '?';
}



function _invoke(op) {
    return new Promise(function (resolve, _) {
        homegear.invoke(op, function (ret) {
            return resolve(ret);
        });
    });
}



function status_role(role_id) {
    return new Promise(function (resolve) {
        const role = interfaceData.roles[role_id];
        const invokes_descs = _status_gather_invokes(role, role_id);

        if ('l2_status' in role) {
            _states.content[role_id] = [];
            return resolve();
        }

        for (const invoke_desc of invokes_descs) {
            const invoke_op = {
                jsonrpc: '2.0',
                method: 'aggregateRoles',
                params: [invoke_desc.type, invoke_desc.ids, [], 0, true],
            };
            homegear.invoke(invoke_op, function (res) {
                /**
                * Updating this is a two step process:
                * 1) We need to fetch the new values from upstream.
                * 2) We need to zero out the outdated old values.
                **/
                const keys = Object.keys(res.result)
                                    .filter(x => x !== 'variableCount');
                for (const key of keys) {
                    /**
                    * We have to be very careful here, to not break Vue's
                    * reactiveness.
                    * 1) In case we need to insert a value, push is needed.
                    * 2) In case we need to change a value, direct
                    *    assignment of the single object members at
                    *    respective array index is needed.
                    **/
                    for (let i = 0; i < _states.content[role_id].length; ++i) {
                        if (_states.content[role_id][i].raw == key) {
                            set_or_extend(_states.content[role_id], i, {
                                key:   _status_text(role, key),
                                value: res.result[key],
                                raw:   key,
                            });
                            break;
                        }
                    }
                }

                for (let j = _states.content[role_id].length - 1; j >= 0; --j) {
                    if (keys.indexOf(_states.content[role_id][j].raw) === -1)
                        _states.content[role_id][j].value = 0;
                }
            });
        }

        return resolve();
    });
}



function states_init() {
    const keys = Object.keys(interfaceData.roles).map(Number);

    for (const key of keys) {
        /**
         * Always reinitialize to an empty array here.
         * In case of a homegear reconnect, we would duplicate the
         * status text otherwise.
         **/
        Vue.set(_states.content, key, []);

        const role = interfaceData.roles[key];
        if (!('texts' in role))
            continue;

        /**
         * This must be a for-in loop! Do not change it into a
         * traditional for loop or a map.
         * We need iterate over both, integer keys, as well as object
         * keys (iteration order does not matter for the integer keys).
         **/
        for (const text_idx in role.texts)
            _states.content[key].push(
                {
                    key:   _status_text(role, text_idx),
                    raw:   text_idx + '',
                    value: 0,
                }
            );
    }

    Promise.allSettled(keys.map(key => status_role(key)))
           .then(() => _states.initialized = true);
}
homegear.ready(states_init);



function states_flag_dirty(roles) {
    _states.dirty = _states.dirty.concat(roles)
                                 .filter((v, i, a) => a.indexOf(v) === i);
}



const mixin_states = {
    computed: {
        states_is_ready: function () {
            return _states.initialized;
        },

        states: function () {
            return _states.content;
        },

        states_is_dirty: function () {
            return _states.dirty.length > 0;
        },
    },

    methods: {
        states_refetch_dirty: function () {
            if (! this.states_is_dirty)
                return;

            const dirty = _states.dirty.slice();
            _states.dirty = [];

            Promise.allSettled(dirty.map(x => status_role(x)));
        },
    },
};

/*
    global
        ModeMenuState
        mixin_menus
        mixin_modemenu
        mixin_notification
        mixin_print_mounted
        mixin_profiles
*/
/*
    exported
        ShifMainmenu
        ShifModemenu
*/



const ShifMainmenu = {
    mixins: [
        mixin_menus,
        mixin_notification,
        mixin_print_mounted()
    ],

    data: function () {
        return {
            enabled_menus: interfaceData.mainmenu
                            .filter(x => ! this.disabled('mainmenu', x.name)),
            active: true,
        };
    },

    computed: {
        width: function () {
            return (100 / this.enabled_menus.length) + '%';
        },

        notifications_available: function () {
            return this.integrated_notifications.length;
        },
    },

    methods: {
        badge_wanted: function (cur) {
            return this.notifications_available > 0 && cur === 'notifications';
        }
    },

    template: `
        <div id="mainmenu">
            <ul class="menu">
                <li v-for="i in enabled_menus"
                    v-bind:style="{width: width}">

                    <router-link v-bind:to="{name: i.name}">
                        <div v-bind:id="'mainmenu_' + i.name"
                             class="mainmenu_button"
                             v-bind:class="{badge: badge_wanted(i.name)}"
                             v-bind:data-badge="notifications_available">
                             <shif-icon v-bind:src="i.icon" />
                        </div>
                    </router-link>

                </li>
            </ul>
        </div>
    `,
};



const ShifModemenuHidden = {
    template: `
        <div></div>
    `,
};



const ShifModemenuDraggable = {
    mixins: [mixin_modemenu],

    template: `
        <div>
            <div class="mode_text">
                <span class="mode_name">{{ i18n('modemenu.sort.name') }}</span>
            </div>
            <div class="mode_buttons">
                <div v-on:click="modemenu_hide" class="mode_end">
                    {{ i18n('modemenu.sort.button.end') }}
                </div>
            </div>
        </div>
    `,
};



const ShifModemenuFavorites = {
    mixins: [mixin_modemenu],

    template: `
        <div>
            <div class="mode_text">
                <span class="mode_name">{{ i18n('modemenu.favorites.name') }}</span>
            </div>
            <div class="mode_buttons">
                <div v-on:click="modemenu_hide" class="mode_end">
                    {{ i18n('modemenu.favorites.button.end') }}
                </div>
            </div>
        </div>
    `,
};



const ShifModemenuProfiles = {
    mixins: [mixin_modemenu, mixin_profiles],

    computed: {
        link_profile: function () {
            return {
                name: 'settings.profiles.profile',
                params: {
                    profile_id: this.$root.profiles.id,
                },
            };
        },
    },

    methods: {
        submit_profile: function () {
            this.profile_update(interfaceData.profiles[this.$root.profiles.id],
                                this.$root.profiles.form);
        },
    },

    template: `
        <div id="mode_wrapper_profiles">
            <div class="mode_text">
                <span class="mode_label">
                    {{ i18n('modemenu.profiles.name.label') }}:
                </span>
                <span class="mode_name">
                    {{ $root.profiles.form.profile_name }}
                </span>
            </div>
            <div class="mode_buttons">
                <router-link v-bind:to="link_profile">
                    <div class="mode_settings">
                        {{ i18n('modemenu.profiles.button.settings') }}
                    </div>
                </router-link>
                <div v-on:click="submit_profile"
                     class="mode_end">
                    {{ i18n('modemenu.profiles.button.end') }}
                </div>
            </div>
        </div>
    `
};



const _modemenu_components = {
    [ModeMenuState.HIDDEN]:    ShifModemenuHidden,
    [ModeMenuState.FAVORITES]: ShifModemenuFavorites,
    [ModeMenuState.PROFILES]:  ShifModemenuProfiles,
    [ModeMenuState.DRAGGABLE]: ShifModemenuDraggable,
};



const ShifModemenu = {
    mixins: [mixin_modemenu],

    computed: {
        content: function () {
            return _modemenu_components[this.modemenu_state];
        },
    },

    template: `
        <div id="modemenu">
            <component v-if="content"
                       v-bind:is="content" />
        </div>
    `,
};

/*
    global
        i18n
*/
/*
    exported
        user_register_webauthn_device
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
// Register a new WebAuthn device
////////////////////////////////////////////////////////////////////////////////////////////////////////
function user_register_webauthn_device() {
    /**
     * convert RFC 1342-like base64 strings to array buffer
     * @param {mixed} obj
     * @returns {undefined}
     */
    function recursiveBase64StrToArrayBuffer(obj) {
        let prefix = '?BINARY?B?';
        let suffix = '?=';
        if (typeof obj === 'object') {
            for (let key in obj) {
                if (typeof obj[key] === 'string') {
                    let str = obj[key];
                    if (str.substring(0, prefix.length) === prefix && str.substring(str.length - suffix.length) === suffix) {
                        str = str.substring(prefix.length, str.length - suffix.length);

                        let binary_string = window.atob(str);
                        let len = binary_string.length;
                        let bytes = new Uint8Array(len);
                        for (var i = 0; i < len; i++)        {
                            bytes[i] = binary_string.charCodeAt(i);
                        }
                        obj[key] = bytes.buffer;
                    }
                } else {
                    recursiveBase64StrToArrayBuffer(obj[key]);
                }
            }
        }
    }

    /**
     * Convert a ArrayBuffer to Base64
     * @param {ArrayBuffer} buffer
     * @returns {String}
     */
    function arrayBufferToBase64(buffer) {
        var binary = '';
        var bytes = new Uint8Array(buffer);
        var len = bytes.byteLength;
        for (var i = 0; i < len; i++) {
            binary += String.fromCharCode( bytes[ i ] );
        }
        return window.btoa(binary);
    }

    if(interfaceData.options.userHasTwofaRegistrations) return;
    if (!window.fetch || !navigator.credentials || !navigator.credentials.create) {
        $('#registerWebauthn').val(i18n('settings.user.manage.browserNotSupported'));
        $('#registerWebauthn').attr('disabled', 'disabled');
        return;
    }

    window.fetch('webauthnServer.php?command=getCredentialArgs', {method:'GET',cache:'no-cache'}).then(function(response) {
        return response.json();
    }).then(function(json) {
        recursiveBase64StrToArrayBuffer(json);
        return json;
    }).then(function(createCredentialArgs) {
        return navigator.credentials.create(createCredentialArgs);
    }).then(function(cred) {
        return {
            clientDataJson: cred.response.clientDataJSON ? arrayBufferToBase64(cred.response.clientDataJSON) : null,
            attestationObject: cred.response.attestationObject ? arrayBufferToBase64(cred.response.attestationObject) : null
        };
    }).then(JSON.stringify).then(function(AuthenticatorAttestationResponse) {
        return window.fetch('webauthnServer.php?command=registerDevice', {method:'POST', body: AuthenticatorAttestationResponse, cache:'no-cache'});
    }).then(function(response) {
        return response.json();
    }).then(function(json) {
        if (json.success) {
            $('#registerWebauthn').val(i18n('settings.user.manage.twofaRegistered'));
            $('#registerWebauthn').attr('disabled', 'disabled');
        } else {
            throw new Error(json.msg);
        }
    }).catch(function(error) {
        $('#registerWebauthn').val(error.message);
        $('#registerWebauthn').attr('disabled', 'disabled');
    });
}

/*
    global
        clone
        homegear
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
        shif_device_slot_profiles
        shif_register_disable_hooks
*/



let shading_windalarm = clone(shif_device);
shading_windalarm.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status_minimal()"
                     v-bind:place="place">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('shadingWindalarm', shading_windalarm, shading_windalarm);
shif_comps_create('shadingRainalarm', shading_windalarm, shading_windalarm);



let shading_ventilation_l2 = clone(shif_device);
shading_ventilation_l2.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, !this.props.value);
};

shading_ventilation_l2.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-on:click_icon="change"
                     v-on:click="level3(device)">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let shading_ventilation_l3 = clone(shif_device);
shading_ventilation_l3.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, !this.props.value);
};

shading_ventilation_l3.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:place="place"
                     v-bind:status="status_minimal()"
                     v-bind:disabled="disabled"
                     v-on:click="change">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('shadingVentilation', shading_ventilation_l2, shading_ventilation_l3);



let shading_slider = clone(shif_device);
shading_slider.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, this.props.value);
};

shading_slider.template = `
    <shif-slider v-bind:min="props.minimumScaled"
                 v-bind:max="props.maximumScaled"
                 v-bind:unit="props.unit"
                 v-bind:value="props.value"
                 v-bind:title="title"
                 v-bind:step=5
                 v-bind:disabled="disabled"
                 v-on:change="change"
                 v-model:value="props.value">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-slider>
`;

shif_comps_create('shadingPosition',      shading_ventilation_l2, shading_slider);
shif_comps_create('shadingPositionSlats', shading_ventilation_l2, shading_slider);



let shading_buttons_l2 = clone(shif_device);
shading_buttons_l2.template = `
    <shif-generic-l2 v-bind:icon="icons.l2.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: icons.l2.color, text: texts.title.color}"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-bind:status="status"
                     v-on:click="level3(device)">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let shading_buttons_l3 = clone(shif_device);
shading_buttons_l3.methods.change = function(_event) {
    let upVar = this.control.variableOutputs[0];
    let downVar = this.control.variableOutputs[2];
    if((this.indexes.input == 0 || this.indexes.input == 2) &&
        !upVar.hasOwnProperty('value') &&
        !downVar.hasOwnProperty('value') &&
        upVar.peer == downVar.peer &&
        upVar.channel == downVar.channel &&
        upVar.name == downVar.name) {
        let down = (this.indexes.input == 2);
        if((this.output.type == 'integer' ||
            this.output.type == 'integer64' ||
            this.output.type == 'float') &&
            this.output.hasOwnProperty('minimumValue') &&
            this.output.hasOwnProperty('maximumValue')) {
            let output = clone(this.output);
            output.value = (down ? this.output.maximumValue : this.output.minimumValue);
            homegear.value_set_clickcounter(this, output, null);
        }
        else if(this.output.type == 'bool') {
            let output = clone(this.output);
            output.value = down;
            homegear.value_set_clickcounter(this, output, null);
        }
    }
    else homegear.value_set_clickcounter(this, this.output, true);
};

shading_buttons_l3.template = `
    <div class="control_button_wrapper">
        <shif-button v-bind:disabled="disabled"
                     v-on:click="change">
            <shif-icon v-bind:src="cond.icon.name"
                        v-bind:active="cond.icon.color">
            </shif-icon>

            ${shif_device_slot_automations}
            ${shif_device_slot_draggable}

        </shif-button>
    </div>
`;

shif_comps_create('shadingButtons', shading_buttons_l2, shading_buttons_l3);
shif_comps_create('shadingButtonsSingleStep', shading_buttons_l2, shading_buttons_l3);
shif_comps_create('shadingButtonsUpDown', shading_buttons_l2, shading_buttons_l3);



shif_register_disable_hooks({
    'Base.shadingButtonsSingleStepPositionPositionSlatsWindalarm': [
        {
            condition: {
                index: 6,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3, 4, 5],
            reason: 'Windalarm',
        },
    ],
});

shif_register_disable_hooks({
    'Base.shadingButtonsPositionPositionSlatsRainalarm': [
        {
            condition: {
                index: 0,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3, 4],
            reason: 'Rainalarm',
        },
    ],
});

shif_register_disable_hooks({
    'Base.shadingButtonsPositionVentilationWindalarm': [
        {
            condition: {
                index: 5,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3, 4],
            reason: 'Windalarm',
        },
    ],
});

shif_register_disable_hooks({
    'Base.shadingButtonsPositionWindalarm': [
        {
            condition: {
                index: 4,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3],
            reason: 'Windalarm',
        },
    ],
});


shif_register_disable_hooks({
    'Base.shadingAwningButtonsPositionRainalarm': [
        {
            condition: {
                index: 5,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3],
            reason: 'Disabled',
        },
        {
            condition: {
                index: 3,
                operator: 'eq',
                value: 1
            },
            disable: [0, 1, 2, 3],
            reason: 'Rainalarm'
        }
    ],
});

shif_register_disable_hooks({
    'Base.shadingAwningButtonsPositionWindalarm': [
        {
            condition: {
                index: 4,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3],
            reason: 'Windalarm'
        }
    ],
});

shif_register_disable_hooks({
    'Base.shadingAwningButtonsPositionWindalarmRainalarm': [
        {
            condition: {
                index: 4,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3],
            reason: 'Windalarm',
        },
        {
            condition: {
                index: 5,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3],
            reason: 'Rainalarm',
        }
    ],
});

/*
    global
        clone
        homegear
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
        shif_device_slot_profiles
        shif_register_disable_hooks
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
let window_rainalarm = clone(shif_device);
window_rainalarm.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status_minimal()"
                     v-bind:place="place">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('windowRainalarm', window_rainalarm, window_rainalarm);



let window_status = clone(shif_device);
window_status.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let window_status_l3 = clone(shif_device);
window_status_l3.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status_minimal()"
                     v-bind:place="place">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('windowContact', window_status, window_status_l3);
shif_comps_create('windowHandle',  window_status, window_status_l3);

let window_buttons_l2 = clone(shif_device);
window_buttons_l2.template = `
    <shif-generic-l2 v-bind:icon="icons.l2.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: icons.l2.color, text: texts.title.color}"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-bind:status="status"
                     v-on:click="level3(device)">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;
let window_buttons_l3 = clone(shif_device);
window_buttons_l3.methods.change = function(_event) {
    let upVar = this.control.variableOutputs[0];
    let downVar = this.control.variableOutputs[2];
    if((this.indexes.input == 0 || this.indexes.input == 2) &&
        !upVar.hasOwnProperty('value') &&
        !downVar.hasOwnProperty('value') &&
        upVar.peer == downVar.peer &&
        upVar.channel == downVar.channel &&
        upVar.name == downVar.name) {
        let down = (this.indexes.input == 2);
        if((this.output.type == 'integer' ||
            this.output.type == 'integer64' ||
            this.output.type == 'float') &&
            this.output.hasOwnProperty('minimumValue') &&
            this.output.hasOwnProperty('maximumValue')) {
            let output = clone(this.output);
            output.value = (down ? this.output.maximumValue : this.output.minimumValue);
            homegear.value_set_clickcounter(this, output, null);
        }
        else if(this.output.type == 'bool') {
            let output = clone(this.output);
            output.value = down;
            homegear.value_set_clickcounter(this, output, null);
        }
    }
    else homegear.value_set_clickcounter(this, this.output, true);
};
window_buttons_l3.template = `
    <div>
        <div class="control_button_wrapper">
            <shif-button v-bind:disabled="disabled"
                         v-on:click="change">
                <shif-icon v-bind:src="cond.icon.name"
                           v-bind:active="cond.icon.color">
                </shif-icon>

                ${shif_device_slot_automations}
                ${shif_device_slot_draggable}

            </shif-button>
        </div>
    </div>
`;

shif_comps_create('windowButtonsUpDown', window_buttons_l2, window_buttons_l3);
shif_comps_create('windowButtons', window_buttons_l2, window_buttons_l3);

let window_slider = clone(shif_device);
window_slider.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, this.props.value);
};
window_slider.provides = function () {
    return {
        checkbox_wanted: true,
    };
},
window_slider.template = `
    <shif-slider v-bind:min="props.minimumScaled"
                 v-bind:max="props.maximumScaled"
                 v-bind:unit="props.unit"
                 v-bind:value="props.value"
                 v-bind:title="title"
                 v-bind:step=5
                 v-bind:disabled="disabled"
                 v-on:change="change"
                 v-model:value="props.value">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-slider>
`;

shif_comps_create('windowPosition',      window_buttons_l2, window_slider);



shif_register_disable_hooks({
    'Base.windowButtonsUpDownRainalarm': [
        {
            condition: {
                index: 2,
                operator: 'eq',
                value: true
            },
            disable: [0, 1],
            reason: 'Rainalarm',
        },
    ],
});

shif_register_disable_hooks({
    'Base.windowButtonsPositionRainalarm': [
        {
            condition: {
                index: 4,
                operator: 'eq',
                value: true
            },
            disable: [0, 1, 2, 3],
            reason: 'Rainalarm',
        },
    ],
});

/*
    global
        clone
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
let door_status = clone(shif_device);
door_status.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status_minimal()"
                     v-bind:place="place">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('doorContact', door_status, door_status);
shif_comps_create('doorHandle',  door_status, door_status);
shif_comps_create('doorLock',  door_status, door_status);

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
    global
        clone
        homegear
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
        shif_device_slot_profiles
        shif_register_disable_hooks
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
let lighting_switch_l2 = clone(shif_device);
lighting_switch_l2.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, !this.props.value);
};
lighting_switch_l2.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-on:click_icon="change($event, true)"
                     v-on:click="level3(device)">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let lighting_switch_l3 = clone(shif_device);
lighting_switch_l3.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, !this.props.value);
};
lighting_switch_l3.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:place="place"
                     v-bind:status="status_minimal()"
                     v-on:click="change">

        ${shif_device_slot_favorites}
        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}


    </shif-generic-l2>
`;

shif_comps_create('lightingSwitch', lighting_switch_l2, lighting_switch_l3);



let lighting_brightness = clone(shif_device);
lighting_brightness.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, this.props.value);
};
lighting_brightness.template = `
    <shif-slider v-bind:min="props.minimumScaled"
                 v-bind:max="props.maximumScaled"
                 v-bind:unit="props.unit"
                 v-bind:value="props.value"
                 v-bind:title="title"
                 v-bind:step=5
                 v-bind:disabled="disabled"
                 v-on:change="change"
                 v-model:value="props.value">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}


    </shif-slider>
`;

shif_comps_create('lightingBrightness', lighting_switch_l2, lighting_brightness);
shif_comps_create('lightingSpeed', lighting_switch_l2, lighting_brightness);



let lighting_button_l2 = clone(shif_device);
lighting_button_l2.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place"
                     v-on:click="level3(device)">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}


    </shif-generic-l2>
`;

let lighting_button_l3 = clone(shif_device);
lighting_button_l3.methods.change = function(_event, down) {
    homegear.value_set_clickcounter(this, this.output, down);
};
lighting_button_l3.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status_minimal()"
                     v-bind:place="place"
                     v-on:mousedown="change($event, true)"
                     v-on:mouseup="change($event, false)">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}


    </shif-generic-l2>
`;

shif_comps_create('lightingButton', lighting_button_l2, lighting_button_l3);
shif_comps_create('refresh', lighting_button_l2, lighting_button_l3);



let lighting_color_l2 = clone(shif_device);
lighting_color_l2.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place"
                     v-on:click="level3(device)">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}


    </shif-generic-l2>
`;

let lighting_color_l3 = clone(shif_device);
lighting_color_l3.methods.change = function(event) {
    this.props.value = event;
    homegear.value_set_clickcounter(this, this.output, this.props.value);
};
lighting_color_l3.template = `
    <shif-colorpicker v-bind:width="{max_pixels: 600, percent: 100}"
                      v-bind:height="{max_pixels: 620, percent: 100}"
                      v-bind:color="props.value"
                      v-bind:title="title"
                      v-on:input="change">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}


    </shif-colorpicker>
`;

shif_comps_create('lightingColor', lighting_color_l2, lighting_color_l3);

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
let lighting_function_l2 = clone(shif_device);
lighting_function_l2.template = `
    <shif-generic-l2 v-bind:icon="icons.l2.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: icons.l2.color, text: texts.title.color}"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-bind:status="status"
                     v-on:click="level3(device)">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}


    </shif-generic-l2>
`;

let lighting_function_l3 = clone(shif_device);
lighting_function_l3.methods.change = function(x) {
    homegear.value_set_clickcounter(this, this.output, x);
};
lighting_function_l3.computed.values = function () {
    return this.rendering
        .map(x => ({
            name:     x.definitions.texts.state.content,
            value:    x.condition.value,
        }));
};
lighting_function_l3.template = `
    <shif-dropdown v-bind:title="title"
                   v-bind:values="values"
                   v-bind:selected="props.value"
                   v-on:change="x => change(parseInt(x))">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}


    </shif-dropdown>
`;

shif_comps_create('lightingFunction', lighting_function_l2, lighting_function_l3);



shif_register_disable_hooks({
    'Base.lightingSwitchButtonBrightnessColorDEMO': [
        {
            condition: {
                or: [
                    {
                        index: 0,
                        operator: 'eq',
                        value: false,
                    },
                    {
                        index: 1,
                        operator: 'eq',
                        value: false,
                    },
                ],
            },
            disable: [3, 4],
            reason: 'Disabled because device is turned off',
        },
    ],
});

/*
    global
        clone
        homegear
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
        shif_device_slot_profiles
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
let socket_switch_l2 = clone(shif_device);
socket_switch_l2.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, !this.props.value);
};
socket_switch_l2.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-on:click_icon="change"
                     v-on:click="level3(device)">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let socket_switch_l3 = clone(shif_device);
socket_switch_l3.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, !this.props.value);
};
socket_switch_l3.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:place="place"
                     v-bind:status="status_minimal()"
                     v-on:click="change">

        ${shif_device_slot_favorites}
        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('socketSwitch', socket_switch_l2, socket_switch_l3);



let socket_button_l2 = clone(shif_device);
socket_button_l2.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place"
                     v-on:click="level3(device)">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let socket_button_l3 = clone(shif_device);
socket_button_l3.methods.change = function(_event, down) {
    homegear.value_set_clickcounter(this, this.output, down);
};
socket_button_l3.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status_minimal()"
                     v-bind:place="place"
                     v-on:mousedown="change($event, true)"
                     v-on:mouseup="change($event, false)">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('socketButton', socket_button_l2, socket_button_l3);

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
    global
        clone
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
function status_format(status, precision) {
    if (typeof(status) === 'string') {
        const regex = /^(\d+(?:[.,]\d+)?)/;
        return status.replace(regex, (match) => this.float_formatted(match, precision));
    }

    for (let cur of status)
        cur.value = status_format.apply(this, [cur.value, precision]);

    return status;
}



let climate_main = clone(shif_device);
climate_main.computed.status_formatted = function () {
    return status_format.apply(this, [this.status_minimal(false), 1]);
};
climate_main.computed.icon = function () {
    return 'main' in this.icons
        ? this.icons.main
        : this.control.icons.main;
};
climate_main.template = `
    <shif-generic-l2 v-bind:icon="icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: icon.color, text: texts.title.color}"
                     v-bind:status="status_formatted"
                     v-bind:place="place">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;



let openweathermap_l2 = clone(shif_device);
openweathermap_l2.computed.status_formatted = function () {
    return status_format.apply(this, [this.status, 1]);
};
openweathermap_l2.template = `
    <shif-generic-l2 v-bind:icon="icons.weather.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: icons.weather.color, text: texts.title.color}"
                     v-bind:status="status_formatted"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-on:click="level3(device)">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;



shif_comps_create('humidity',             openweathermap_l2, climate_main);
shif_comps_create('temperature',          openweathermap_l2, climate_main);
shif_comps_create('climateCity',          openweathermap_l2, climate_main);
shif_comps_create('climateCloudcoverage', openweathermap_l2, climate_main);
shif_comps_create('climateDate',          openweathermap_l2, climate_main);
shif_comps_create('climatePressure',      openweathermap_l2, climate_main);
shif_comps_create('climateRain3h',        openweathermap_l2, climate_main);
shif_comps_create('climateSnow3h',        openweathermap_l2, climate_main);
shif_comps_create('climateSunrise',       openweathermap_l2, climate_main);
shif_comps_create('climateSunset',        openweathermap_l2, climate_main);
shif_comps_create('climateWeather',       openweathermap_l2, climate_main);
shif_comps_create('climateWinddirection', openweathermap_l2, climate_main);
shif_comps_create('climateWindgust',      openweathermap_l2, climate_main);
shif_comps_create('climateWindspeed',     openweathermap_l2, climate_main);
// shif_comps_create('OpenWeatherMap',    openweathermap_l2, openweathermap_l2);

/*
    global
        clone
        homegear
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
        shif_device_slot_profiles
        shif_register_disable_hooks
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
let heating_slider_l2 = clone(shif_device);
heating_slider_l2.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, !this.props.value);
};
heating_slider_l2.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-on:click_icon="change"
                     v-on:click="level3(device)">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;
let heating_slider_l3 = clone(shif_device);
heating_slider_l3.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, this.props.value);
};
heating_slider_l3.template = `
    <shif-slider v-bind:min="props.minimumScaled"
                 v-bind:max="parseFloat(props.maximumScaled)"
                 v-bind:unit="props.unit"
                 v-bind:value="props.value"
                 v-bind:title="title"
                 v-bind:step="0.5"
                 v-bind:precision="1"
                 v-on:change="change"
                 v-bind:disabled="disabled"
                 v-model:value="props.value">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-slider>
`;

shif_comps_create('heatingSlider', heating_slider_l2, heating_slider_l3);



let heating_is_state_l2 = clone(shif_device);
heating_is_state_l2.template = `
    <shif-generic-l2 v-bind:icon="control.icons.temperature.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: control.icons.temperature.color, text: texts.title.color}"
                     v-bind:status="status"
                     v-bind:actions="true"
                     v-bind:place="place"
                     v-on:click="level3(device)">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;
let heating_is_state_l3 = clone(shif_device);
heating_is_state_l3.template = `
    <shif-generic-l2 v-bind:icon="control.icons.temperature.name"
                     v-bind:title="title"
                     v-bind:active="{icon: control.icons.temperature.color, text: texts.title.color}"
                     v-bind:status="status"
                     v-bind:place="place">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('heatingIsState', heating_is_state_l2, heating_is_state_l3);



let heating_mode_l3 = clone(shif_device);
heating_mode_l3.computed.values = function () {
    return this.rendering
        .map(x => ({
            name:     x.definitions.texts.state.content,
            value:    x.condition.value,
            selected: x.condition.value == this.props.value,
        }));
};
heating_mode_l3.methods.change = function(x) {
    homegear.value_set_clickcounter(this, this.output, x);
};
heating_mode_l3.template = `
    <shif-radio v-bind:title="title"
                v-bind:values="values"
                v-bind:disabled="disabled"
                v-on:input="x => change(parseInt(x))">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-radio>
`;

shif_comps_create('heatingMode', heating_is_state_l2, heating_mode_l3);



let heating_window = clone(shif_device);
heating_window.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="title"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status_minimal()"
                     v-bind:place="place">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

shif_comps_create('heatingWindow', heating_window, heating_window);



shif_register_disable_hooks({
    'Base.heatingIsStateSliderMode': [
        {
            condition: {
                index: 2,
                operator: 'eq',
                value: 0
            },
            disable: [1],
            reason: 'Frost',
        }
    ],
});

shif_register_disable_hooks({
    'Base.heatingIsStateSliderModeWindow': [
        {
            condition: {
                index: 2,
                operator: 'eq',
                value: 0
            },
            disable: [1],
            reason: 'Frost'
        },
        {
            condition: {
                index: 3,
                operator: 'gt',
                value: 0
            },
            disable: [1, 2],
            reason: 'Window',
        }
    ],
});

shif_register_disable_hooks({
    'Base.heatingIsStateSliderModeWindowHandle': [
        {
            condition: {
                index: 2,
                operator: 'eq',
                value: 0
            },
            disable: [1],
            reason: 'Frost'
        },
        {
            condition: {
                index: 3,
                operator: 'gt',
                value: 0
            },
            disable: [1, 2],
            reason: 'Window',
        }
    ],
});

shif_register_disable_hooks({
    'Base.heatingIsStateSliderModeWindowContact': [
        {
            condition: {
                index: 2,
                operator: 'eq',
                value: 0
            },
            disable: [1],
            reason: 'Frost'
        },
        {
            condition: {
                index: 3,
                operator: 'gt',
                value: 0
            },
            disable: [1, 2],
            reason: 'Window',
        }
    ],
});

/*
    global
        clone
        homegear
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
        shif_device_slot_profiles
*/



////////////////////////////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
let ventilation_mode_l2 = clone(shif_device);
ventilation_mode_l2.template = `
    <shif-generic-l2 v-bind:icon="icons.l2.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: icons.l2.color, text: texts.title.color}"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-bind:status="status"
                     v-on:click="level3(device)">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let ventilation_mode_l3 = clone(shif_device);
ventilation_mode_l3.methods.change = function(x) {
    homegear.value_set_clickcounter(this, this.output, x);
};
ventilation_mode_l3.computed.values = function () {
    return this.rendering
        .map(x => ({
            name:     x.definitions.texts.state.content,
            value:    x.condition.value,
            selected: x.condition.value == this.props.value,
        }));
};
ventilation_mode_l3.template = `
    <shif-radio v-bind:title="title"
                v-bind:values="values"
                v-on:input="x => change(parseInt(x))">

        ${shif_device_slot_profiles}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-radio>
`;

shif_comps_create('ventilationMode', ventilation_mode_l2, ventilation_mode_l3);

/*
    global
        clone
        homegear
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
*/



let timer_l2 = clone(shif_device);
timer_l2.template = `
    <shif-generic-l2 v-bind:icon="cond.icon.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: cond.icon.color, text: cond.text.color}"
                     v-bind:status="status"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-on:click="level3(device)">

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let timer_l3 = clone(shif_device);
timer_l3.methods.change = function(_event) {
    homegear.value_set_clickcounter(this, this.output, !this.props.value);
};

timer_l3.template = `
    <div class="control_wrapper">
        <div class="device">
            <div class="checkbox_wrapper">
                <div class="checkbox_automation_wrapper">
                <router-link v-if="used_by_automations !== false"
                                v-bind:to="automation_link">
                    <shif-icon src="calendar-time_1" />
                </router-link>
                </div>
            </div>
            <shif-title v-bind:disabled="disabled">Zeitschaltung</shif-title>
            <form>
                <div class="timer_wrapper">
                    <div>
                        <input type="time" name="time">
                    </div>
                    <div>
                        <select name="sun">
                            <option value="false">-</option>
                            <option value="sa">Sonnenaufgang</option>
                            <option value="su">Sonnenuntergang</option>
                        </select>
                    </div>
                    <div>
                        <select name="state">
                            <option value="false">off</option>
                            <option value="true">on</option>
                        </select>
                    </div>
                    <div style="max-width: 456px;">
                        <select class="select_day" name="day" multiple>
                            <option value="1">Mo.</option>
                            <option value="2">Di.</option>
                            <option value="3">Mi.</option>
                            <option value="4">Do.</option>
                            <option value="5">Fr.</option>
                            <option value="6">Sa.</option>
                            <option value="7">So.</option>
                        </select>
                    </div>
                    <div>
                        <select name="activ">
                            <option value="false">inactive</option>
                            <option value="true">active</option>
                        </select>
                    </div>
                    <div>
                        <input type="submit" name="submit" value="Hinzufügen">
                    </div>
                </div>
            </form>
        </div>
    </div>
`;

shif_comps_create('timer', timer_l2, timer_l3);

/*
    global
        clone
        homegear
        shif_comps_create
        shif_device
        shif_device_slot_automations
        shif_device_slot_draggable
        shif_device_slot_favorites
*/



let media_buttons_l2 = clone(shif_device);
media_buttons_l2.template = `
    <shif-generic-l2 v-bind:icon="icons.l2.name"
                     v-bind:title="dev.label"
                     v-bind:active="{icon: icons.l2.color, text: texts.title.color}"
                     v-bind:place="place"
                     v-bind:actions="true"
                     v-bind:status="status"
                     v-on:click="level3(device)">

        ${shif_device_slot_favorites}
        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </shif-generic-l2>
`;

let media_buttons_l3 = clone(shif_device);
media_buttons_l3.methods.change = function(_event) {
    let upVar = this.control.variableOutputs[0];
    let downVar = this.control.variableOutputs[2];
    if((this.indexes.input == 0 || this.indexes.input == 2) &&
        !upVar.hasOwnProperty('value') &&
        !downVar.hasOwnProperty('value') &&
        upVar.peer == downVar.peer &&
        upVar.channel == downVar.channel &&
        upVar.name == downVar.name) {
        let down = (this.indexes.input == 2);
        if((this.output.type == 'integer' ||
            this.output.type == 'integer64' ||
            this.output.type == 'float') &&
            this.output.hasOwnProperty('minimumValue') &&
            this.output.hasOwnProperty('maximumValue')) {
            let output = clone(this.output);
            output.value = (down ? this.output.maximumValue : this.output.minimumValue);
            homegear.value_set_clickcounter(this, output, null);
        }
        else if(this.output.type == 'bool') {
            let output = clone(this.output);
            output.value = down;
            homegear.value_set_clickcounter(this, output, null);
        }
    }
    else homegear.value_set_clickcounter(this, this.output, true);
};

media_buttons_l3.template = `
    <div class="control_button_wrapper">
        <shif-button v-bind:disabled="disabled"
                     v-on:click="change">
            <shif-icon v-bind:src="cond.icon.name"
                        v-bind:active="cond.icon.color">
            </shif-icon>
        </shif-button>

        ${shif_device_slot_automations}
        ${shif_device_slot_draggable}

    </div>
`;

shif_comps_create('mediaButtons', media_buttons_l2, media_buttons_l3);

/*
    global
        check_disabled
        get_or_default
        mixin_components
        mixin_modemenu
        mixin_profiles
        mixin_print_mounted
        mixin_scroll_position
        mixin_states
        shif_device_slot_draggable
*/
/*
    exported
        ShifAllDevicesLvl3
        ShifHouseDevices
        ShifHouseDevicesRole
*/



Vue.component('shif-ctrl-summary', {
    mixins: [mixin_components, mixin_modemenu, mixin_profiles, mixin_print_mounted()],

    props: [
        'icon',
        'title',
        'status',
        'actions',
        'devs',
        'role_id',
    ],

    data: function () {
        return {
            submenu_show: this.role_id === this.last_role_id,
        };
    },

    computed: {
        dev_objs: function () {
            return this.devs.map(x => this.find_component(this.interfaceData.devices[x], 'l2'));
        },

        profiles_by_role: function () {
            return this.role_profiles(this.role_id);
        },
    },

    provide: function () {
        return {
            layer: 2,
            role_id: this.role_id,
        };
    },

    inject: {
        last_role_id: {from: 'role_id'},
    },

    watch: {
        submenu_show: function () {
            this.$emit('accordion-open', this.role_id);
        },
    },

    methods: {
        toggle_all: function(action) {
            const varInRole = 'roleId' in action
                            ? this.interfaceData.roles[action.roleId].varInRole
                            : this.interfaceData.roles[this.role_id].varInRole;

            let ops = [];
            for (const peer in varInRole)
                for (const channel in varInRole[peer])
                    for (const name in varInRole[peer][channel]) {
                        const cur = varInRole[peer][channel][name];

                        if ('direction' in cur && cur.direction === 0)
                            continue;

                        if (! (peer    in this.interfaceData.map_invoke &&
                               channel in this.interfaceData.map_invoke[peer] &&
                               name    in this.interfaceData.map_invoke[peer][channel]))
                            continue;
                        const devs = this.interfaceData.map_invoke[peer][channel][name];

                        const disabled = devs.some(dev => {
                            const device  = this.interfaceData.devices[dev.databaseId];
                            const control = device.controls[dev.control];

                            for (const i in control.variableInputs) {
                                if (check_disabled(device, {control: dev.control, input: i}).flag)
                                    return true;
                            }

                            return false;
                        });


                        if (! disabled)
                            ops.push({
                                input: {peer, channel, name},
                                value: action.value
                            });
                    }

            this.$homegear.value_set_multi(ops);
        },

        get_icon_or_default: function (profile) {
            return get_or_default(profile, 'icon', 'slider_1');
        },
    },

    template: `
        <div>
            <shif-generic-l2 v-bind:icon="icon"
                             v-bind:icon_rotate="submenu_show"
                             v-bind:title="title"
                             v-bind:status="status"
                             v-bind:actions="true"
                             v-bind:accordion="true"
                             v-on:click="submenu_show = !submenu_show"
                             class="accordion">
                ${shif_device_slot_draggable}
            </shif-generic-l2>

            <shif-trans-drop-down>
                <div v-if="submenu_show"
                     class="categoryContainer"
                     style="margin-top: 15px;">

                    <div class="control_button_wrapper">
                        <template v-for="action in actions">
                            <shif-button v-bind:width="(100 / actions.length) + '%'"
                                         v-on:click="toggle_all(action)">
                                {{ action.buttonText }}
                            </shif-button>
                        </template>
                    </div>

                    <div class="profiles_wrapper">
                        <template v-for="i in profiles_by_role">
                            <shif-generic-l2 v-bind:icon="get_icon_or_default(i)"
                                             v-bind:title="i.name"
                                             v-bind:status="i18n('modemenu.profiles.name.label')"
                                             v-bind:active="{icon: i.isActive ? 'active' : ''}"
                                             v-on:click="profile_start(i)">
                            </shif-generic-l2>
                        </template>
                    </div>

                    <shif-draggable v-bind:value="dev_objs"
                                    v-slot="draggable"
                                    v-bind:name="role_id"
                                    handle=".drag_drop_icon">
                        <template v-for="dev in draggable.values">
                            <component v-bind="dev.val" v-bind:include_place="true" />
                        </template>
                    </shif-draggable>
                </div>
            </shif-trans-drop-down>
        </div>
    `,
});

                            // <template v-if="$root.debug">
                                // {{ dev | pretty | log }}
                            // </template>



let ShifAllDevices = {
    mixins: [mixin_modemenu, mixin_states, mixin_print_mounted()],

    data: function () {
        return {
            role_id_opened: null,
        };
    },

    computed: {
        map_roles_devs: function () {
            let ret = {};

            for (const dev_idx in interfaceData.devices) {
                let dev = interfaceData.devices[dev_idx];
                if (!('role' in dev))
                    continue;

                const role = dev.role;
                if (!(role in ret))
                    ret[role] = [];

                if (ret[role].indexOf(dev_idx) === -1)
                    ret[role].push(dev_idx);
            }

            return ret;
        },
    },

    created: function () {
        this.states_refetch_dirty();
    },

    watch: {
        states_is_dirty: function () {
            this.states_refetch_dirty();
        },
    },

    template: `
        <div>
            <shif-draggable v-bind:value="map_roles_devs"
                            v-slot="draggable"
                            handle=".drag_drop_icon">
                <template v-for="i in draggable.values">

                    <template v-if="i.key in interfaceData.roles">
                        <shif-ctrl-summary
                            v-bind:actions="interfaceData.roles[i.key].invokeOutputs"
                            v-bind:icon="interfaceData.roles[i.key].icon"
                            v-bind:title="interfaceData.roles[i.key].name"
                            v-bind:devs="i.val"
                            v-bind:role_id="i.key"
                            v-bind:status="states[i.key]"
                            v-on:accordion-open="x => role_id_opened = x"
                            >
                        </shif-ctrl-summary>
                    </template>

                    <template v-else>
                        {{ "This role is not defined: " + i.key | log }}
                    </template>

                </template>
            </shif-draggable>
        </div>
    `,
};


let ShifHouseDevices = {
    mixins: [mixin_scroll_position, mixin_print_mounted('shif-house-devices')],

    components: {
        ShifAllDevices,
    },

    props: {
        room_id:   { },
        device_id: { },
        floor_id:  { },
        role_id:   { },
    },

    provide: function () {
        return {
            room_id:   this.room_id,
            floor_id:  this.floor_id,
            device_id: this.device_id,
            role_id:   this.role_id,
        };
    },

    template: `
        <shif-mainmenu-tabs>
            <shif-all-devices ref="devices" />
        </shif-mainmenu-tabs>
    `
};



let ShifHouseDevicesRole = {
    mixins: [mixin_scroll_position, mixin_print_mounted('shif-house-devices')],

    components: {
        ShifAllDevices,
    },

    props: {
        role_id:   { },
        room_id:   { },
        device_id: { },
        floor_id:  { },
    },

    provide: function () {
        return {
            role_id:   this.role_id,
            room_id:   this.room_id,
            floor_id:  this.floor_id,
            device_id: this.device_id,
        };
    },

    template: `
        <shif-mainmenu-tabs>
            <shif-all-devices />
        </shif-mainmenu-tabs>
    `
};



let ShifAllDevicesLvl3 = {
    mixins: [mixin_scroll_position, mixin_print_mounted('shif-all-devices-level-3')],

    props: {
        room_id:   { required: true, },
        device_id: { required: true, },
        floor_id:  { required: true, },
        role_id:   { required: true, },
    },

    provide: function () {
        return {
            room_id:   this.room_id,
            floor_id:  this.floor_id,
            device_id: this.device_id,
            role_id:   this.role_id,
        };
    },

    template: `
        <shif-house-collected-entries v-bind:layer="3" />
    `
};

/*
    global
        mixin_print_mounted
        mixin_profiles
        mixin_scroll_position
*/
/*
    exported
        ShifFavorites
        ShifFavoritesLvl1
        ShifFavoritesLvl3
*/



let ShifFavoritesLvl3 = {
    mixins: [mixin_print_mounted('shif-favorites-level-3')],

    props: {
        room_id:   { required: true, },
        device_id: { },
        floor_id:  { required: true, },
    },

    provide: function () {
        return {
            floor_id:  this.floor_id,
            device_id: this.device_id,
            room_id:   this.room_id,
        };
    },

    template: `
        <shif-house-collected-entries v-bind:layer="3" />
    `
};



let ShifFavoritesLvl1 = {
    mixins: [
        mixin_scroll_position,
        mixin_profiles,
        mixin_print_mounted('shif-favorites-level-1')
    ],

    props: {
        room_id:   { },
        floor_id:  { },
        device_id: { },
    },

    provide: function () {
        return {
            floor_id:  this.floor_id,
            device_id: this.device_id,
            room_id:   this.room_id,
        };
    },

    template: `
        <shif-house-collected-entries v-bind:layer="2"
                                    v-bind:favorites="true"
                                    v-bind:include_place="true" />
    `,
};



let ShifFavorites = {
    mixins: [mixin_print_mounted('shif-favorites')],

    template: `
        <shif-paging id="favorites" />
    `
};

/*
    global
        draggable_create
        mixin_modemenu
        mixin_rooms
        mixin_print_mounted
        mixin_scroll_position
*/
/*
    exported
        ShifHouse
        ShifHouseLvl2
        ShifHouseLvl3
        ShifHouseRooms
*/



Vue.component('shif-house-floors-rooms', {
    mixins: [mixin_modemenu, mixin_rooms, mixin_print_mounted()],

    data: function () {
        const windowWidth = $( window ).width() - 100;
        const roomSelectWrapperWidth = 185 + 30;
        let maxWidth = 1080;

        const rooms_per_floor = Object.keys(interfaceData.floors)
                                      .map(x => interfaceData.floors[x].rooms.length);
        const max_rooms_per_floor = Math.max.apply(this, rooms_per_floor);

        let roomSelectWrapperMaxCount = windowWidth / roomSelectWrapperWidth;
        roomSelectWrapperMaxCount = roomSelectWrapperMaxCount.toString().split('.')[0];

        if (max_rooms_per_floor > roomSelectWrapperMaxCount)
            maxWidth = roomSelectWrapperMaxCount * roomSelectWrapperWidth + 20;
        else
            maxWidth = max_rooms_per_floor * roomSelectWrapperWidth + 25;

        return {
            maxWidth: maxWidth + 'px',
        };
    },

    computed: {
        has_multiple_floors: function () {
            return Object.keys(interfaceData.floors).length > 1;
        },

        rooms_per_floor: function () {
            let out = {};

            for (const i in interfaceData.floors) {
                const cur = interfaceData.floors[i];
                out[i] = draggable_create(cur.rooms);
            }

            return out;
        }
    },

    template: `
        <div id="house_rooms" v-bind:style="{'max-width': maxWidth}">
            <shif-draggable v-bind:value="interfaceData.floors"
                            v-slot="draggable_floor"
                            handle=".drag_drop_icon">
                <div v-for="floor in draggable_floor.values">
                    <div v-if="has_multiple_floors || unassigned_rooms.length > 0"
                         class="roomSelectTitle"
                         >
                        <shif-icon v-if="modemenu_is_state('DRAGGABLE')"
                                   src="move_1" classname="drag_drop_icon" />
                        {{ floor.val.name }}
                    </div>

                    <div class="rooms_wrapper">
                        <shif-draggable v-bind:value="floor.val.rooms"
                                        v-bind:name="floor.key"
                                        v-slot="draggable_room"
                                        handle=".drag_drop_icon_right">
                            <shif-room v-for="room in draggable_room.values"
                                       v-if="room_has_devices(room.val)"
                                       v-bind:key="room.val"
                                       v-bind:floor="{key: floor.key, value: floor.val}"
                                       v-bind:room="room.val" />
                        </shif-draggable>
                    </div>
                </div>
            </shif-draggable>

            <template v-if="unassigned_rooms.length > 0">
                <div class="roomSelectTitle">
                    {{ i18n('house.storyless') }}
                </div>

                <div class="rooms_wrapper">
                    <shif-room v-for="room_val in unassigned_rooms"
                               v-bind:key="room_val"
                               v-bind:floor="{key: -1}"
                               v-bind:room="room_val"
                               />
                </div>
            </template>
        </div>
    `,
});



let ShifHouseLvl3 = {
    mixins: [mixin_print_mounted('shif-house-lvl3')],

    props: {
        room_id:   { required: true, },
        device_id: { required: true, },
        floor_id:  { required: true, },
    },

    provide: function () {
        return {
            room_id:   this.room_id,
            floor_id:  this.floor_id,
            device_id: this.device_id,
        };
    },

    template: `
        <shif-house-collected-entries v-bind:layer="3" />
    `
};



let ShifHouseLvl2 = {
    mixins: [mixin_scroll_position, mixin_print_mounted('shif-house-lvl2')],

    props: {
        room_id:   { required: true, },
        device_id: { },
        floor_id:  { required: true, },
    },

    provide: function () {
        return {
            floor_id:  this.floor_id,
            device_id: this.device_id,
            room_id:   this.room_id,
        };
    },

    template: `
        <shif-house-collected-entries v-bind:layer="2" />
    `
};



let ShifHouseRooms = {
    mixins: [mixin_scroll_position, mixin_print_mounted('shif-house-rooms')],

    template: `
        <shif-mainmenu-tabs>
            <shif-house-floors-rooms />
        </shif-mainmenu-tabs>
    `
};



let ShifHouse = {
    mixins: [mixin_print_mounted('shif-house')],

    template: `
        <shif-paging id="house" />
    `,
};

/*
    global
        viewLog
*/
/*
    exported
        ShifLog
*/



let ShifLog = Vue.component('shif-log', {
    mounted: function () {
        viewLog('-----------------------------');
    },

    template: `
        <pre id="log" class="content content_single">
        </pre>
    `
});

/*
    global
        NotificationDisplayType
        mixin_notification
        mixin_print_mounted
        mixin_scroll_position
*/
/*
    exported
        ShifNotifications
        ShifNotificationsLvl1
        ShifNotificationsNotification
        modal_mount
*/



// @vue/component
const ShifNotificationMessage = {
    props: {
        modalTitle: {
            type: String,
            default: undefined,
        },
        modalContent: {
            type: String,
            default: undefined,
        },
        buttons: {
            type: Array,
            default: function () {
                return [];
            },
        },
    },

    computed: {
        has_buttons: function () {
            return this.buttons.length > 0;
        },
    },

    methods: {
        on_click: function (btn) {
            this.$emit('click', {
                button: btn
            });
        },
    },

    template: `
        <div class="notification_message">
            <div class="notification_title" v-if="modalTitle">{{ modalTitle }}</div>

            <div class="notification_content" v-if="modalContent" v-html="modalContent">
            </div>

            <div class="notification_buttons">
                <template v-if="has_buttons">
                    <button v-for="i in buttons"
                            v-bind:class="i.type"
                            v-on:click="on_click(i)">
                        <shif-icon v-if="i.icon !== undefined"
                                   v-bind:src="i.icon" />
                        <span v-html="i.label"></span>
                    </button>
                </template>
                <template v-else>
                    <button v-on:click="$emit('close_delete')">
                        {{ i18n('notifications.close_delete') }}
                    </button>
                </template>
            </div>
        </div>
    `
};



Vue.component('shif-notifications-element', {
    mixins: [mixin_print_mounted()],

    props: {
        name:        { type: String, required: true, },
        description: { type: String, default: undefined, },
        icon:        { type: String, default: 'notification_1'},
        translate:   { type: Boolean, default: true, },
    },

    computed: {
        icon_active_class: function () {
            return this.icon_active ? 'active' : '';
        },
    },

    template: `
        <div class="button">
            <shif-icon v-if="icon" classname="button_icon" v-bind:src="icon" />
            <div class="button_text">
                <template v-if="description && description.length > 0">
                    <div class="button_title">{{ translate ? i18n(name) : name }}</div>
                </template>
                <template v-else="description && description.length > 0">
                    <div class="button_title button_no_description">{{ translate ? i18n(name) : name }}</div>
                </template>
                <template v-if="description && description.length > 0">
                    <div class="button_status" style="display:block;">{{ translate ? i18n(description) : description }}</div>
                </template>
            </div>
            <div class="button_action">
                <svg xmlns="http://www.w3.org/2000/svg" version="1.1" id="svg" x="0" y="0" width="370.81" height="370.81" viewBox="0 0 370.81 370.81">
                    <g id="Ebene_1">
                        <path d="M77.9 345.97L102.03 370.81 292.92 185.41 102.03 0 77.9 24.85 243.18 185.41z"/>
                    </g>
                </svg>
            </div>
        </div>
    `
});



let modal = new Vue({
    name: 'Modal',

    components: {
        ShifNotificationMessage,
    },

    mixins: [mixin_notification],

    computed: {
        show: function () {
            return this.modal_notifications.length > 0;
        },
    },

    methods: {
        click: function (id, _btn) {
            const btn = _btn.button;

            this.notification_action(id, btn.id, () => {
                const msg = interfaceData.notifications[id];

                if (msg.displayType === NotificationDisplayType.Modal &&
                    btn.closeModal === true)
                    this.notification_hide(id);

                if (btn.reloadUi === true)
                    window.location.reload();
            });
        },

        close_delete: function (id) {
            this.notification_remove(id);
        },
    },

    template: `
        <div v-if="show" id="modal">
            <div class="content content_single">
                <template v-for="i in modal_notifications">
                    <shif-notification-message v-bind:key="i.id"
                                               v-bind="i"
                                               v-on:click="click(i.id, $event)"
                                               v-on:close_delete="close_delete(i.id)" />
                </template>
            </div>
        </div>
    `
});



// @vue/component
const ShifNotificationsNotification = {
    components: {
        ShifNotificationMessage,
    },

    mixins: [
        mixin_notification,
        mixin_print_mounted('shif-notifications-notification')
    ],

    props: {
        notification_id: {}
    },

    computed: {
        msg: function () {
            return interfaceData.notifications[this.notification_id];
        },
    },

    watch: {
        'msg': function (new_) {
            if (new_)
                return;

            this.$router.back();
        },
    },

    methods: {
        click: function (_btn) {
            const btn = _btn.button;

            this.notification_action(this.notification_id, btn.id, () => {
                const msg = interfaceData.notifications[this.notification_id];

                if (msg.displayType === NotificationDisplayType.Modal &&
                    btn.closeModal === true)
                    this.notification_hide(msg.id);

                if (btn.reloadUi === true)
                    window.location.reload();
            });
        },

        close_delete: function () {
            this.notification_remove(this.notification_id);
        },
    },

    template: `
        <div>
            <shif-notification-message v-bind="msg"
                                       v-on:close_delete="close_delete"
                                       v-on:click="click" />
        </div>
    `
};



// @vue/component
const ShifNotificationsLvl1 = {
    mixins: [
        mixin_notification,
        mixin_scroll_position,
        mixin_print_mounted('shif-notifications-lvl1')
    ],

    methods: {
        link: function (id) {
            return {
                name: 'notifications.notification',
                params: {
                    notification_id: id,
                },
            };
        },
    },

    template: `
        <div>
            <template v-for="i in integrated_notifications">
                <router-link v-bind:to="link(i.id)">
                    <shif-notifications-element v-bind:key="i.id"
                                                v-bind:name="i.title"
                                                v-bind:translate="false"
                                                icon="notification_1" />
                </router-link>
            </template>
        </div>
    `
};



// @vue/component
const ShifNotifications = {
    mixins: [mixin_print_mounted('shif-notifications')],

    template: `
        <shif-paging id="notifications" />
    `
};



function modal_mount() {
    modal.$mount('#modal');
}

/*
    global
        mixin_print_mounted
        mixin_profiles
*/
/*
    exported
        ShifProfiles
*/



let ShifProfilesGlobal = {
    mixins: [mixin_profiles, mixin_print_mounted('shif-profiles-global')],

    provide: {
        layer: 2,
    },

    template: `
        <div id="profiles" class="content content_single">
            <div class="profiles_wrapper">
                <template v-for="i in global_profiles">
                    <shif-generic-l2 v-bind:icon="profile_icon(i)"
                                     v-bind:title="i.name"
                                     v-bind:active="{icon: i.isActive ? 'active' : ''}"
                                     v-on:click="profile_start(i)">
                        <template v-slot:automations>
                            <router-link v-if="profile_used_by_automations(i.id) !== false"
                                         v-bind:to="profile_automation_link(i.id)">
                                <shif-icon src="calendar-time_1" />
                            </router-link>
                        </template>
                    </shif-generic-l2>
                </template>
            </div>
        </div>
    `
};



let ShifProfiles = {
    mixins: [mixin_print_mounted('shif-profiles')],

    components: {
        ShifProfilesGlobal,
    },

    template: `
        <shif-mainmenu-tabs>
            <shif-profiles-global />
        </shif-mainmenu-tabs>
    `
};

/*
    global
        ModeMenuState
        i18n
        licenses
        mixin_menus
        mixin_modemenu
        mixin_print_mounted
        mixin_profiles
        mixin_rooms
        mixin_scroll_position
        user_register_webauthn_device
*/
/*
    exported
        ShifSettings
        ShifSettingsAutomations
        ShifSettingsAutomationsForm
        ShifSettingsFavorites
        ShifSettingsItems
        ShifSettingsLicenses
        ShifSettingsProfile
        ShifSettingsProfiles
        ShifSettingsSort
        ShifSettingsUser
*/



Vue.component('shif-settings-element', {
    mixins: [mixin_print_mounted()],

    props: {
        name:        { type: String, required: true, },
        description: { type: String, },
        icon:        { type: String, default: 'slider_1'},
        icon_active: { type: Boolean, default: false, },
        translate:   { type: Boolean, default: true, }
    },

    computed: {
        icon_active_class: function () {
            return this.icon_active ? 'active' : '';
        },
    },

    template: `
        <div class="button">
            <shif-icon v-if="icon" classname="button_icon" v-bind:src="icon" v-bind:active="icon_active_class" />
            <div class="button_text">
                <template v-if="description && description.length > 0">
                    <div class="button_title">{{ translate ? i18n(name) : name }}</div>
                </template>
                <template v-else="description && description.length > 0">
                    <div class="button_title button_no_description">{{ translate ? i18n(name) : name }}</div>
                </template>
                <template v-if="description && description.length > 0">
                    <div class="button_status" style="display:block;">{{ translate ? i18n(description) : description }}</div>
                </template>
            </div>
            <div class="button_action">
                <svg xmlns="http://www.w3.org/2000/svg" version="1.1" id="svg" x="0" y="0" width="370.81" height="370.81" viewBox="0 0 370.81 370.81">
                    <g id="Ebene_1">
                        <path d="M77.9 345.97L102.03 370.81 292.92 185.41 102.03 0 77.9 24.85 243.18 185.41z"/>
                    </g>
                </svg>
            </div>
        </div>
    `
});



let ShifSettingsLicenses = {
    mixins: [mixin_print_mounted('shif-settings-licenses')],

    computed: {
        data: function () {
            return licenses;
        },
    },

    methods: {
        log_hook: function () {
            this.$router.push('/log');
        },
    },

    template: `
        <div>
        <div class="table1">
            <table>
                <thead>
                    <th v-on:click="log_hook">{{ i18n('settings.about.table.name') }}</th>
                    <th>{{ i18n('settings.about.table.version') }}</th>
                    <th>{{ i18n('settings.about.table.rights') }}</th>
                    <th>{{ i18n('settings.about.table.license') }}</th>
                    <th>{{ i18n('settings.about.table.license.url') }}</th>
                </thead>
                <tbody>
                    <tr>
                        <td>{{ interfaceData.manifest.name }}</td>
                        <td>{{ interfaceData.manifest.revision }}</td>
                        <td>{{ interfaceData.manifest.author }}</td>
                        <td>{{ interfaceData.manifest.license }}</td>
                        <td>{{ interfaceData.manifest.licenseurl }}</td>
                    </tr>
                    <tr v-for="i in data">
                        <td>{{ i.name }}</td>
                        <td>{{ i.version }}</td>
                        <td>{{ i.rights }}</td>
                        <td>{{ i.licensename }}</td>
                        <td>{{ i.licenseurl }}</td>
                    </tr>
                </tbody>
            </table>
        </div>
        </div>
    `
};



let ShifSettingsUser = {
    mixins: [mixin_print_mounted('shif-settings-user')],

    data: function () {
        return {
            options:   interfaceData.options,
            startpaths: [
                'favorites',
                'house.tab.rooms',
                'house.tab.devices',
                'house.tab.profiles',
            ],
            languages: interfaceData.i18n.languages,
            name:      this.$route.name,

            form: {
                name: 'user_edit',
                startPath: interfaceData.options.startPath,
                language:  interfaceData.options.language,
                theme:     interfaceData.options.theme,
                color:     interfaceData.options.highlight,
            }
        };
    },

    computed: {
        themes: function () {
            return interfaceData.themes.filter(x => x.name !== 'standard' &&
                                                    x.name !== 'highlight');
        },

        two_fa_value: function () {
            return this.options.userHasTwofaRegistrations
                ? i18n(this.name + '.twofaRegistered')
                : i18n(this.name + '.registerTwofa');
        },
    },

    methods: {
        form_submit: function () {
            this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'getUserMetadata',
                params: []
            }, (data) => {
                let new_settings = data.result;

                new_settings.locale    = this.form.language;

                if (new_settings.interface === undefined)
                    new_settings.interface = {};

                Object.assign(new_settings.interface, {
                    language:  this.form.language  || this.options.language  || 'en-US',
                    startPath: this.form.startPath || this.options.startPath || 'house.tab.rooms',
                    theme:     this.form.theme     || 'dark',
                    highlight: this.form.color     || '#e3a104',
                });

                this.$homegear.invoke({
                    jsonrpc: '2.0',
                    method: 'setUserMetadata',
                    params: [new_settings]
                }, function () {
                    window.location.reload(true);
                });
            });
        },

        two_fa_register_dev: function () {
            return user_register_webauthn_device();
        },
    },

    template: `
        <div class="user_wrapper">
            <form v-bind:id="form.name"
                  v-bind:name="form.name"
                  action="javascript:void(0);"
                  v-on:submit.stop="form_submit">

                <template v-if="options.twofaChangeable && options.twofaEnabled">
                    <div class="form-group">
                        <div class="label">{{ i18n(name + '.twofa') }}:</div>
                        <input id="registerWebauthn"
                               type="button"
                               v-bind:disabled="options.userHasTwofaRegistrations"
                               v-bind:value="two_fa_value"
                               v-on:click="two_fa_register_dev" />
                    </div>
                </template>

                <template v-if="options.languageChangeable">
                    <div class="form-group">
                        <div class="label">{{ i18n(name + '.language') }}:</div>
                        <select id="language"
                                name="language"
                                v-model="form.language">
                            <option v-for="value, key in languages"
                                    v-bind:value="key"
                                    autocomplete="off">
                                {{ value.name }}
                            </option>
                        </select>
                    </div>
                </template>

                <div class="form-group">
                    <div class="label">{{ i18n(name + '.startpath') }}:</div>
                    <select id="startpath"
                            name="startpath"
                            v-model="form.startPath">
                        <option v-for="i in startpaths"
                                v-bind:value="i"
                                autocomplete="off">
                            {{ i18n(i) }}
                        </option>
                    </select>
                </div>

                <div class="form-group">
                    <div class="label">{{ i18n(name + '.theme') }}:</div>
                    <select id="theme"
                            name="theme"
                            v-model="form.theme">
                        <option v-for="i in themes"
                                v-bind:value="i.name"
                                autocomplete="off">
                                {{ i18n(name + '.theme.' + i.name) }}
                        </option>
                    </select>
                </div>

                <div id="user_highlight_display" class="form-group">
                    <div class="label">{{ i18n(name + '.highlight') }}:</div>
                    <input id="highlight"
                           type="hidden"
                           name="highlight"
                           v-bind:value="options.highlight" />
                    <shif-colorpicker v-bind:width=" {max_pixels: 600, percent: 100}"
                                      v-bind:height="{max_pixels: 620, percent: 100}"
                                      v-bind:color="options.highlight"
                                      v-model="form.color"
                                      />
                    <div style="clear:both;"></div>
                </div>

                <div class="form-group">
                    <input type="submit"
                           v-bind:name="form.name"
                           v-bind:value="i18n(name + '.save')" />
                </div>

            </form>
        </div>
    `
};



let ShifSettingsItems = function (level) {
    return {
        mixins: [
            mixin_scroll_position,
            mixin_menus,
            mixin_print_mounted('shif-settings-items')
        ],

        methods: {
            elements: function (level) {
                return interfaceData.menu.filter(x => x.mainmenu === 'settings' &&
                                                      x.level    === String(level) &&
                                                      ! this.disabled('menu', x.name));
            },
        },

        template: `
            <div>
                <template v-for="i in elements(${level})">
                    <router-link v-bind:to="{name: i.name}">
                        <shif-settings-element v-bind:key="i.name"
                                               v-bind:icon="i.icon"
                                               v-bind:name="i.name"
                                               v-bind:description="i.description" />
                    </router-link>
                </template>
            </div>
        `,
    };
};



let ShifSettingsSort = {
    mixins: [mixin_modemenu],

    data: function () {
        return {
            state: this.modemenu_is_state(ModeMenuState.DRAGGABLE),
        };
    },

    watch: {
        'state': function () {
            if (!this.state)
                return this.modemenu_hide();

            this.modemenu_show(ModeMenuState.DRAGGABLE);
            this.$router.push({name: 'house.tab.rooms'});
        },
    },

    template: `
        <div class="device_wrapper">
            <div class="device"
                 v-on:click.prevent="state = ! state">
                <shif-title>{{ i18n('settings.sort.mode') }}</shif-title>
                <shif-checkbox v-model="state" />
            </div>
        </div>
    `
};



let ShifSettingsFavorites = {
    mixins: [mixin_modemenu],

    data: function () {
        return {
            state: this.modemenu_is_state(ModeMenuState.FAVORITES),
        };
    },

    watch: {
        'state': function () {
            if (!this.state)
                return this.modemenu_hide();

            this.modemenu_show(ModeMenuState.FAVORITES);
            this.$router.push({name: 'house.tab.rooms'});
        },
    },

    template: `
        <div class="device_wrapper">
            <div class="device"
                 v-on:click.prevent="state = ! state">
                <shif-title>{{ i18n('settings.favorites.mode') }}</shif-title>
                <shif-checkbox v-model="state" />
            </div>
        </div>
    `
};



let ShifSettingsProfiles = {
    mixins: [
        mixin_scroll_position,
        mixin_profiles,
        mixin_print_mounted('shif-settings-profile')
    ],

    methods: {
        link: function (profile_id) {
            return {
                name: 'settings.profiles.profile',
                params: {
                    profile_id: profile_id,
                },
            };
        },
    },

    template: `
        <div>
            <div class="profiles_wrapper">
                <template v-for="i in editable_profiles">
                    <router-link v-bind:to="link(i.id)">
                        <shif-settings-element v-bind:key="i.name"
                                               v-bind:name="i.name"
                                               v-bind:translate="false"
                                               v-bind:icon="i.icon"
                                               description=""
                                               />
                    </router-link>
                </template>

                <router-link v-bind:to="{name: 'settings.profiles.new'}">
                    <shif-settings-element name="settings.profiles.new"
                                           icon="plus_1" />
                </router-link>
            </div>
        </div>
    `
};



let ShifSettingsProfileRoleValue = {
    mixins: [mixin_print_mounted('shif-settings-profile-role-value')],

    props: {
        value: {
            required: true,
        },
        role_id: {
            required: true,
        },
    },

    data: function () {
        return {
            value_: this.value,
        };
    },

    computed: {
        role: function () {
            return interfaceData.roles[this.role_id];
        },

        role_values: function () {
            return Object.keys(this.role.roleProfileValues.options)
                         .map(x => ({id: x, name: this.role.roleProfileValues.options[x]}))
                         .concat({id: null, name: '---'});
        },
    },

    watch: {
        value_: function () {
            this.$emit('input', this.value_);
            this.$emit('change', this.value_);
        },
    },

    template: `
        <div>
            <select v-if="role.roleProfileValues.type === 'select'"
                    v-model="value_">
                <option v-for="i in role_values"
                        v-bind:value="i.id">
                    {{ i.name }}
                </option>
            </select>
        </div>
    `,
};



let ShifSettingsProfile = {
    mixins: [
        mixin_modemenu,
        mixin_profiles,
        mixin_print_mounted('shif-settings-profile')
    ],

    props: [
        'profile_id',
    ],

    components: {
        ShifSettingsProfileRoleValue,
    },

    data: function () {
        if (this.profile_id === undefined ||
            ! (this.profile_id in interfaceData.profiles))
            return {
                mode: 'add',
                profile: null,
                form: {
                    name: 'profile_add',
                    icon: 'slider_1',
                    profile_name: '',
                    role: {
                        role: null,
                        value: null,
                    },
                    location: {
                        floor:    null,
                        room:     null,
                        global:   false,
                        favorite: false,
                    },
                }
            };

        const profile = interfaceData.profiles[this.profile_id];

        if (this.is_enabled) {
            return {
                mode: 'edit',
                profile: profile,
                form: this.$root.profiles.form,
            };
        }

        const [floor, room] = profile.locations.length === 0
                ? [null, null]
                : [profile.locations[0].floorId, profile.locations[0].roomId];

        const global   = profile.global === true;
        const favorite = profile.favorite === true;

        const role = profile.roles !== undefined && profile.roles.length > 0
                        ? profile.roles[0]
                        : {role: null, value: null};

        return {
            mode: 'edit',
            profile: profile,
            form: {
                name: 'profile_edit',
                icon: profile.icon,
                profile_name: profile.name,
                location: {
                    floor:    floor === undefined ? null : floor,
                    room:     room  === undefined ? null : room,
                    global:   global,
                    favorite: favorite,
                },
                role: role,
            },
        };
    },

    computed: {
        is_enabled: function () {
            return this.modemenu_is_state(ModeMenuState.PROFILES);
        },

        floors: function () {
            return Object.keys(interfaceData.floors)
                         .map(x => ({id: x, name: interfaceData.floors[x].name}))
                         .concat({id: null, name: '---'});
        },

        filtered_rooms: function () {
            let rooms = Object.keys(interfaceData.rooms);

            if (this.form.location.floor !== undefined &&
                this.form.location.floor !== null &&
                this.form.location.floor in interfaceData.floors)
                rooms = interfaceData.floors[this.form.location.floor]
                                     .rooms;

            return rooms.map(x => ({id: x, name: interfaceData.rooms[x].name}))
                        .concat({id: null, name: '---'});
        },

        filtered_roles: function () {
            return Object.keys(interfaceData.roles)
                         .filter(x => interfaceData.roles[x].roleProfileValues !== undefined)
                         .map(x => ({id: x, name: interfaceData.roles[x].name}))
                         .concat({id: null, name: '---'});
        },

        show_roles: function () {
            return interfaceData.options.roleProfileDefinable === true;
        },
    },

    methods: {
        form_submit: function (source) {
            switch (source) {
                case 'load':
                    this.$root.profiles.form = this.form;
                    this.$root.profiles.id   = this.profile.id;

                    return this.profile_load(this.profile,
                        () => this.$router.push({name: 'house.tab.rooms'})
                    );

                case 'save':
                    if (this.mode === 'edit')
                        return this.profile_update(this.profile, this.form);

                    return this.profile_add(this.form,
                        (result) => this.$router.replace({
                            name: 'settings.profiles.profile',
                            params: {
                                profile_id: result.result
                            },
                        })
                    );

                case 'delete':
                    return this.profile_delete(this.profile,
                        () => this.$router.replace({name: 'settings.profiles'})
                    );
            }
        },
    },

    mounted: function () {
        if (this.mode === 'edit' && ! this.is_enabled)
            return this.profile_build_root_devs(this.profile);

        this.modemenu_hide();
    },

    template: `
        <div class="profiles_wrapper">
            <form v-bind:id="form.name"
                  v-bind:name="form.name"
                  action="javascript:void(0);"
                  v-on:submit.stop="">

                <div class="form-group">
                    <div class="label">{{ i18n('settings.profiles.profile.name') }}:</div>
                    <input id="profile_name"
                           type="text"
                           name="profile_name"
                           v-model="form.profile_name" />
                </div>

                <div class="form-group">
                    <div class="label">{{ i18n('settings.profiles.profile.icon') }}:</div>
                    <shif-icon-selection v-model="form.icon" />
                </div>

                <div class="form-group">
                    <div class="label">{{ i18n('settings.profiles.profile.locations') }}:</div>
                    <div class="global"
                         v-on:click.prevent="form.location.global = !form.location.global">
                        <div class="label">{{ i18n('settings.profiles.profile.locations.global') }}:</div>
                        <shif-checkbox v-model="form.location.global" />
                    </div>
                    <div class="global"
                         v-on:click.prevent="form.location.favorite = !form.location.favorite">
                        <div class="label">{{ i18n('settings.profiles.profile.locations.favorite') }}:</div>
                        <shif-checkbox v-model="form.location.favorite" />
                    </div>
                    <div class="label">{{ i18n('settings.profiles.profile.locations.floor') }}:</div>
                    <select id="locationsFloors"
                            name="locationsFloors"
                            v-model="form.location.floor">
                        <option v-for="i in floors"
                                v-bind:value="i.id"
                                autocomplete="off">
                            {{ i.name }}
                        </option>
                    </select>

                    <div class="label">{{ i18n('settings.profiles.profile.locations.room') }}:</div>
                    <select id="locationsRooms"
                            name="locationsRooms"
                            v-model="form.location.room">
                        <option v-for="i in filtered_rooms"
                                v-bind:value="i.id"
                                autocomplete="off">
                            {{ i.name }}
                        </option>
                    </select>
                </div>

                <div v-if="show_roles"
                     class="form-group">
                    <div class="label">{{ i18n('settings.profiles.profile.roles') }}:</div>

                    <select v-model="form.role.role">
                        <option v-for="i in filtered_roles"
                                v-bind:value="i.id"
                                autocomplete="off">
                            {{ i.name }}
                        </option>
                    </select>

                    <shif-settings-profile-role-value
                        v-if="form.role.role"
                        v-model="form.role.value"
                        v-bind:role_id="form.role.role" />
                </div>

                <div class="form-group">
                    <input v-if="mode === 'edit' && form.role.role === null"
                           type="submit"
                           name="load"
                           v-on:click="form_submit('load')"
                           v-bind:value="i18n('settings.profiles.profile.load')" />

                    <input type="submit"
                           name="save"
                           v-on:click="form_submit('save')"
                           v-bind:value="i18n('settings.profiles.profile.save')" />

                    <input v-if="mode === 'edit'"
                           type="submit"
                           name="delete"
                           v-on:click="form_submit('delete')"
                           v-bind:value="i18n('settings.profiles.profile.delete')" />
                </div>
            </form>
        </div>
    `,
};




const mixin_names = {
    methods: {
        floor_name: function (id) {
            if (id === '-1')
                return i18n('house.storyless');

            return interfaceData.floors[id].name;
        },

        room_name: function (id) {
            return interfaceData.rooms[id].name;
        },

        position: function (floor_id, room_id) {
            return `${this.floor_name(floor_id)} - ${this.room_name(room_id)}`;
        },

        device_name: function (id) {
            return interfaceData.devices[id].label;
        },

        control_name: function (device_id, control_idx) {
            const cur = interfaceData.devices[device_id].controls[control_idx];

            return cur.texts !== undefined &&
                   cur.texts.title !== undefined &&
                   cur.texts.title.content !== undefined
                        ? cur.texts.title.content
                        : cur.uniqueUiElementId;
        },

        input_name: function (device_id, control_idx, input_idx) {
            return interfaceData.devices[device_id]
                                .controls[control_idx]
                                .variableInputs[input_idx]
                                .name
                                .toLowerCase();

        },
    },
};



const _mixin_settings_automations = {
    methods: {
        _automations_add_update: function (data, automation_id, cb) {
            const method = automation_id === undefined
                ? 'addVariableAutomation'
                : 'updateVariableAutomation';
            let params = automation_id === undefined
                ? []
                : [automation_id];

            params.push({
                [interfaceData.options.language]: data.name,
            });

            params.push({
                icon:        data.icon,
                description: data.description,
                conditon:    data.condition,
                action:      data.action,
            });

            return this.$homegear.invoke({
                jsonrpc: '2.0',
                method: method,
                params: params,
            }, (result) => {
                interfaceData.automations[result.result] = {
                    name: data.name,
                    icon: data.icon,
                    description: data.description,
                    condition: Object.assign({}, data.condition),
                    action: Object.assign({}, data.action),
                };

                if (cb)
                    return cb(result);
            });
        },

        automations_add: function (data, cb) {
            return this._automations_add_update(data, undefined, cb);
        },

        automations_update: function (data, automation_id, cb) {
            return this._automations_add_update(data, automation_id, cb);
        },

        automations_delete: function (automation_id, cb) {
            return this.$homegear.invoke({
                jsonrpc: '2.0',
                method: 'deleteVariableAutomation',
                params: [automation_id],
            }, (result) => {
                delete interfaceData.automations[automation_id];

                if (cb)
                    return cb(result);
            });
        },
    },
};



const ShifDeviceSelection = {
    mixins: [
        mixin_names,
        mixin_print_mounted('shif-device-selection')
    ],

    props: {
        map:   { type: Object, required: true, },
        value: { type: Object, },
    },

    methods: {
        option_value: function (floor_id, room_id, device_id) {
            return `${floor_id}-${room_id}-${device_id}`;
        },

        on_input: function (new_) {
            const [floor_id, room_id, device_id] = new_.split('-');

            this.$emit('input', {
                floor_id: floor_id,
                room_id: room_id,
                device_id: device_id,
            });
        },
    },

    template: `
        <select v-bind:value="option_value(value.floor_id, value.room_id, value.device_id)"
                v-on:input="on_input($event.target.value)">
            <option value="null-null-null" disabled>---</option>
            <template v-for="floor, floor_id in map">
                <optgroup v-for="room, room_id in floor"
                          v-bind:label="position(floor_id, room_id)">
                    <option v-for="device_id in room"
                            v-bind:value="option_value(floor_id, room_id, device_id)">
                        {{ device_name(device_id) }}
                    </option>
                </optgroup>
            </template>
        </select>

    `
};



const ShifSettingsAutomationsForm = {
    mixins: [
        mixin_names,
        mixin_rooms,
        mixin_print_mounted('shif-settings-automations-form')
    ],

    components: {
        ShifDeviceSelection,
    },

    props: {
        automation_id: {},
    },

    data: function () {
        const cur = interfaceData.automations[this.automation_id] !== undefined
                        ? interfaceData.automations[this.automation_id]
                        : {condition: {}, action: {}, name: '', description: '', icon: 'calendar-time_1'};

        let astrofunction = cur.condition && cur.condition.astrofunction
            ? Object.assign({}, cur.condition.astrofunction)
            : {trigger: 'sunrise', offset: 0, enabled: false};

        let time = cur.condition && cur.condition.time
            ? Object.assign({}, cur.condition.time)
            : {hour: 0, minute: 0, enabled: false};

        let days = cur.condition && cur.condition.days
            ? Object.assign({}, cur.condition.days)
            : {values: [], enabled: false};

        let condition_devices = cur.condition && cur.condition.devices
            ? {
                enabled: cur.condition.devices.enabled,
                values: cur.condition.devices.values.map(x => Object.assign({}, x))
              }
            : {values: [], enabled: false};

        let action_device = cur.action && cur.action.device
            ? {
                enabled: cur.action.device.enabled,
                value: cur.action.device.value,
                device: Object.assign({}, cur.condition.devices.device)
              }
            : {device: {device_id: null, floor_id: null, room_id: null}, value: '', enabled: false};

        let profile = cur.action && cur.action.profile
            ? Object.assign({}, cur.action.profile)
            : {profile_id: null, enabled: false};

        return {
            when: false,
            then: false,

            name: cur.name,
            icon: cur.icon,
            description: cur.description,
            enabled: cur.enabled === true,

            condition: {
                astrofunction: astrofunction,
                time: time,
                days: days,
                devices: condition_devices,
            },
            action: {
                device: action_device,
                profile: profile,
            },

            form: {
                operators: {
                    eq: '=',
                    ne: '≠',
                    ge: '≥',
                    le: '≤',
                    gt: '>',
                    lt: '<',
                },

                cond_device: {
                    device: { device_id: null, floor_id: null, room_id: null, },
                    control: null,
                    input: null,
                    operator: null,
                    value: '',
                },

                action_device: {
                    device: { device_id: null, floor_id: null, room_id: null, },
                    control: null,
                    input: null,
                    operator: null,
                    value: '',
                },
            },
        };
    },

    methods: {
        i18n: function (key) {
            return i18n('settings.automations.automation.' + key);
        },

        cond_device_add: function () {
            this.condition.devices.values.push({
                floor_id:  this.form.cond_device.device.floor_id,
                room_id:   this.form.cond_device.device.room_id,
                device_id: this.form.cond_device.device.device_id,
                control:   this.form.cond_device.control,
                input:     this.form.cond_device.input,
                operator:  this.form.cond_device.operator,
                value:     this.form.cond_device.value,
            });
        },

        cond_device_delete: function (idx) {
            this.condition.devices.values.splice(idx, 1);
        },
    },

    computed: {
        map_room_floor_device: function () {
            function handle_room(floor_id, room_id) {
                const room = interfaceData.rooms[room_id];

                for (const device_id of room.devices) {
                    if (out[floor_id] === undefined)
                        out[floor_id] = {
                            [room_id]: [device_id],
                        };
                    else if (out[floor_id][room_id] === undefined)
                        out[floor_id][room_id] = [device_id];
                    else if (out[floor_id][room_id].indexOf(device_id) === -1)
                        out[floor_id][room_id].push(device_id);
                }
            }

            let out = {};

            for (const floor_id in interfaceData.floors) {
                const floor = interfaceData.floors[floor_id];

                for (const room_id of floor.rooms)
                    handle_room(floor_id, room_id);
            }

            for (const room_id of this.unassigned_rooms)
                handle_room(-1, room_id);

            return out;
        },

        condition_device: function () {
            if (! this.form.cond_device.device)
                return undefined;

            return interfaceData.devices[this.form.cond_device.device.device_id];
        },

        condition_device_control: function () {
            if (this.condition_device === undefined)
                return undefined;

            return this.condition_device.controls[this.form.cond_device.control];
        },

        action_device: function () {
            if (! this.form.action_device.device)
                return undefined;

            return interfaceData.devices[this.form.action_device.device.device_id];
        },

        action_device_control: function () {
            if (this.action_device === undefined)
                return undefined;

            return this.action_device.controls[this.form.action_device.control];
        },
    },

    template: `
        <div id="automation">
            <div class="form-group automation-state">
                <div class="label">{{ i18n('state') }}:</div>
                <shif-checkbox v-model="enabled" />
            </div>

            <div class="form-group">
                <div class="label">{{ i18n('name') }}:</div>
                <input type="text" v-model="name"/>

                <div class="label">{{ i18n('description') }}:</div>
                <input type="text" v-model="description"/>
            </div>
            <div class="form-group">
                <div class="label">{{ i18n('icon') }}:</div>
                <shif-icon-selection v-model="icon" />
            </div>

            <div class="collapse-content">
                <div class="collapse">
                    <div class="collapse-title" v-on:click="when = !when">
                        <shif-icon classname="collapse-icon" src="slider_1" />
                        <span>{{ i18n('condition') }}</span>
                        <div class="collapse-arrow">
                            <shif-icon src="arrow_down_1" v-bind:rotate="when" />
                        </div>
                    </div>
                    <shif-trans-drop-down>
                        <div v-if="when" class="collapse-content-inner">
                        <div class="inner-content">
                            <div class="timer_wrapper">
                            <div>
                                <div class="form-group collapse-wrapper">
                                    <div class="label">{{ i18n('condition.astrofunction') }}:</div>
                                    <shif-checkbox v-model="condition.astrofunction.enabled" />
                                </div>
                                <template v-if="condition.astrofunction.enabled">
                                    <div class="form-group">
                                        <select name="astro"
                                                v-model="condition.astrofunction.trigger">
                                            <option value="sunrise">
                                                {{ i18n('condition.astrofunction.sunrise') }}
                                            </option>
                                            <option value="sunset">
                                                {{ i18n('condition.astrofunction.sunset') }}
                                            </option>
                                        </select>
                                    </div>
                                    <div class="form-group">
                                        <div class="label">{{ i18n('condition.astrofunction.offset') }}:</div>
                                        <input type="number" name="offset"
                                            v-model.number="condition.astrofunction.offset"/>
                                    </div>
                                </template>
                            </div>
                            </div>

                            <div class="timer_wrapper">
                            <div>
                                <div class="form-group collapse-wrapper">
                                    <div class="label">{{ i18n('condition.time') }}:</div>
                                    <shif-checkbox v-model="condition.time.enabled" />
                                </div>
                                <template v-if="condition.time.enabled">
                                    <div class="form-group time">
                                    <div>
                                        <div class="label">{{ i18n('condition.time.hour') }}:</div>
                                        <input type="number" name="time" min="0" max="23" step="1"
                                            v-model.number="condition.time.hour" />
                                    </div>
                                    <div>
                                        <div class="label">{{ i18n('condition.time.minute') }}:</div>
                                        <input type="number" name="time" min="0" max="59" step="1"
                                            v-model.number="condition.time.minute" />
                                    </div>
                                    </div>
                                </template>
                            </div>
                            </div>

                            <div class="timer_wrapper">
                            <div>
                                <div class="form-group collapse-wrapper">
                                    <div class="label">{{ i18n('condition.days') }}:</div>
                                    <shif-checkbox v-model="condition.days.enabled" />
                                </div>
                                <template v-if="condition.days.enabled">
                                    <select class="select_day" name="days" multiple
                                            v-model.number="condition.days.values">
                                        <option value="1">{{ i18n('condition.days.mo') }}</option>
                                        <option value="2">{{ i18n('condition.days.tu') }}</option>
                                        <option value="3">{{ i18n('condition.days.we') }}</option>
                                        <option value="4">{{ i18n('condition.days.th') }}</option>
                                        <option value="5">{{ i18n('condition.days.fr') }}</option>
                                        <option value="6">{{ i18n('condition.days.sa') }}</option>
                                        <option value="7">{{ i18n('condition.days.su') }}</option>
                                    </select>
                                </template>
                            </div>
                            </div>

                            <div class="timer_wrapper">
                            <div>
                                <div class="form-group collapse-wrapper">
                                    <div class="label">{{ i18n('condition.devices') }}:</div>
                                    <shif-checkbox v-model="condition.devices.enabled" />
                                </div>
                                <template v-if="condition.devices.enabled">
                                    <table class="table1">
                                        <thead>
                                            <tr>
                                                <th>{{ i18n('condition.devices.table.place') }}</th>
                                                <th>{{ i18n('condition.devices.table.device') }}</th>
                                                <th>{{ i18n('condition.devices.table.control') }}</th>
                                                <th>{{ i18n('condition.devices.table.input') }}</th>
                                                <th>{{ i18n('condition.devices.table.operator') }}</th>
                                                <th>{{ i18n('condition.devices.table.value') }}</th>
                                                <th>{{ i18n('condition.devices.table.delete') }}</th>
                                            </tr>
                                        </thead>
                                        <tbody>
                                            <tr v-for="i, idx in condition.devices.values">
                                                <td>{{ position(i.floor_id, i.room_id) }}</td>
                                                <td>{{ device_name(i.device_id) }}</td>
                                                <td>{{ control_name(i.device_id, i.control) }}</td>
                                                <td>{{ input_name(i.device_id, i.control, i.input) }}</td>
                                                <td>{{ form.operators[i.operator] }}</td>
                                                <td>{{ i.value }}</td>
                                                <td>
                                                    <button v-on:click="cond_device_delete(idx)">
                                                        <shif-icon src="close_x_1" />
                                                    </button>
                                                </td>
                                            </tr>
                                        </tbody>
                                    </table>

                                    <div class="automation-device">
                                        <div class="label">{{ i18n('condition.devices.table.device') }}:</div>
                                        <shif-device-selection v-bind:map="map_room_floor_device"
                                                               v-model="form.cond_device.device" />

                                        <div class="label" v-if="condition_device">{{ i18n('condition.devices.table.control') }}:</div>
                                        <select v-if="condition_device"
                                                v-model="form.cond_device.control">
                                            <option v-bind:value="null" disabled>---</option>
                                            <option v-for="i, key in condition_device.controls"
                                                    v-bind:value="key">
                                                {{ control_name(form.cond_device.device.device_id, key) }}
                                            </option>
                                        </select>

                                        <div class="label" v-if="condition_device_control">{{ i18n('condition.devices.table.input') }}:</div>
                                        <select v-if="condition_device_control"
                                                v-model="form.cond_device.input">
                                            <option v-bind:value="null" disabled>---</option>
                                            <option v-for="i, key in condition_device_control.variableInputs"
                                                    v-bind:value="key">
                                                {{ input_name(form.cond_device.device.device_id, form.cond_device.control, key) }}
                                            </option>
                                        </select>

                                        <div class="label" v-if="condition_device_control">{{ i18n('condition.devices.table.operator') }}:</div>
                                        <select v-if="condition_device_control"
                                                v-model="form.cond_device.operator">
                                            <option v-bind:value="null" disabled>---</option>
                                            <option v-for="i, key in form.operators"
                                                    v-bind:value="key">
                                                {{ i }}
                                            </option>
                                        </select>

                                        <div class="label" v-if="condition_device_control">{{ i18n('condition.devices.table.value') }}:</div>
                                        <input v-if="condition_device_control"
                                            v-model.number="form.cond_device.value"
                                            type="text" />
                                    </div>

                                    <button v-on:click="cond_device_add" style="margin-top: 20px">
                                        {{ i18n('condition.devices.add') }}
                                    </button>
                                </template>
                            </div>
                            </div>
                        </div>
                        </div>
                    </shif-trans-drop-down>
                </div>

                <div class="collapse">
                    <div class="collapse-title" v-on:click="then = !then">
                        <shif-icon classname="collapse-icon" src="play_1" />
                        <span>{{ i18n('action') }}</span>
                        <div class="collapse-arrow">
                            <shif-icon src="arrow_down_1" v-bind:rotate="then" />
                        </div>
                    </div>
                    <shif-trans-drop-down>
                        <div v-if="then" class="collapse-content-inner">
                        <div class="inner-content">
                            <div class="timer_wrapper">
                            <div class="automation-device">
                                <div class="form-group collapse-wrapper">
                                    <div class="label">{{ i18n('action.device') }}:</div>
                                    <shif-checkbox v-model="action.device.enabled" />
                                </div>
                                <template v-if="action.device.enabled">
                                    <div class="automation-device">
                                        <div class="label">{{ i18n('condition.devices.table.device') }}:</div>
                                        <shif-device-selection v-bind:map="map_room_floor_device"
                                                               v-model="form.action_device.device" />

                                        <div class="label" v-if="action_device">{{ i18n('condition.devices.table.control') }}:</div>
                                        <select v-if="action_device"
                                                v-model="form.action_device.control">
                                            <option v-bind:value="null" disabled>---</option>
                                            <option v-for="i, key in action_device.controls"
                                                    v-bind:value="key">
                                                {{ control_name(form.action_device.device.device_id, key) }}
                                            </option>
                                        </select>

                                        <div class="label" v-if="action_device_control">{{ i18n('condition.devices.table.input') }}:</div>
                                        <select v-if="action_device_control"
                                                v-model="form.action_device.input">
                                            <option v-bind:value="null" disabled>---</option>
                                            <option v-for="i, key in action_device_control.variableInputs"
                                                    v-bind:value="key">
                                                {{ input_name(form.action_device.device.device_id, form.action_device.control, key) }}
                                            </option>
                                        </select>

                                        <div class="label" v-if="action_device_control">{{ i18n('condition.devices.table.operator') }}:</div>
                                        <select v-if="action_device_control"
                                                v-model="form.action_device.operator">
                                            <option v-bind:value="null" disabled>---</option>
                                            <option v-for="i, key in form.operators"
                                                    v-bind:value="key">
                                                {{ i }}
                                            </option>
                                        </select>

                                        <div class="label" v-if="action_device_control">{{ i18n('condition.devices.table.value') }}:</div>
                                        <input v-if="action_device_control"
                                            v-model.number="form.action_device.value"
                                            type="text" />
                                    </div>
                                </template>
                            </div>
                            </div>

                            <div class="timer_wrapper">
                            <div>
                                <div class="form-group collapse-wrapper">
                                    <div class="label">{{ i18n('action.profile') }}:</div>
                                    <shif-checkbox v-model="action.profile.enabled" />
                                </div>
                                <template v-if="action.profile.enabled">
                                    <select v-model.number="action.profile.profile_id">
                                        <option v-bind:value="null">---</option>
                                        <option v-for="profile, id in interfaceData.profiles"
                                                v-bind:value="id"
                                                v-bind:key="id">
                                            {{ profile.name }}
                                        </option>
                                    </select>
                                </template>
                            </div>
                            </div>
                        </div>
                        </div>
                    </shif-trans-drop-down>
                </div>

            </div>

            <div class="form-group">
                <input type="submit" name="submit" v-bind:value="i18n('save')">
                <input type="submit" name="delete" v-bind:value="i18n('delete')">
            </div>

        </div>
    `,
};



const ShifSettingsAutomations = {
    mixins: [
        mixin_scroll_position,
        mixin_print_mounted('shif-settings-automations')
    ],

    props: {
        automation_ids: {type: String, default: '',},
    },

    methods: {
        link: function (id) {
            return {
                name: 'settings.automations.automation',
                params: {
                    automation_id: id,
                },
            };
        },
    },

    computed: {
        ids: function () {
            return this.automation_ids === ''
                    ? Object.keys(interfaceData.automations)
                    : this.automation_ids.split('-');
        },

        elements: function () {
            return this.ids.map(x => ({
                key: x,
                val: interfaceData.automations[x],
            }));
        }
    },

    template: `
        <div>
            <template v-for="i in elements">
                <router-link v-bind:to="link(i.key)">
                    <shif-settings-element v-bind:key="i.key"
                                            v-bind:description="i.val.description"
                                            v-bind:name="i.val.name"
                                            v-bind:icon="i.val.icon"
                                            v-bind:icon_active="i.val.enabled"
                                            v-bind:translate="false" />
                </router-link>
            </template>

            <router-link v-bind:to="{name: 'settings.automations.new'}">
                <shif-settings-element v-bind:name="'settings.automations.new'"
                                        icon="plus_1" />
            </router-link>
        </div>
    `
};






let ShifSettings = {
    mixins: [mixin_print_mounted('shif-settings')],

    template: `
        <shif-paging id="settings" />
    `,
};

/*
    global
        NotificationDisplayType
        ShifHouse
        ShifHouseRooms
        ShifHouseLvl2
        ShifHouseLvl3
        ShifHouseDevices
        ShifAllDevicesLvl3
        ShifProfiles
        ShifSettings
        ShifSettingsAutomations
        ShifSettingsAutomationsForm
        ShifSettingsFavorites
        ShifSettingsItems
        ShifSettingsLicenses
        ShifSettingsProfile
        ShifSettingsProfiles
        ShifSettingsSort
        ShifSettingsUser
        ShifFavorites
        ShifFavoritesLvl1
        ShifFavoritesLvl3
        ShifNotifications
        ShifNotificationsLvl1
        ShifNotificationsNotification
        ShifLog
        ShifMainmenu
        ShifModemenu

        condition_check
        date_format
        homegear
        i18n
        icons
        mixin_modemenu
        mixin_notification
        user_logoff
*/
/*
    exported
        check_disabled_backend
        check_disabled_frontend
        toast
        interface_mount
        scroll_positions
*/



function check_disabled_backend(device, indexes) {
    function check_event_trigger(event) {
        if (event.trigger == undefined || event.trigger.length != 3)
            return false;

        const trigger = event.trigger;
        return trigger[0] in interfaceData.map_invoke &&
               trigger[1] in interfaceData.map_invoke[trigger[0]] &&
               trigger[2] in interfaceData.map_invoke[trigger[0]][trigger[1]];
    }

    function check_event_disable(event, control_idx) {
        return 'disable' in event &&
               control_idx in event.disable;
    }

    function check_event_condition(event) {
        return 'condition' in event &&
               'operator'  in event.condition &&
               'value'     in event.condition;
    }

    const ret_enabled = {flag: false};

    if (! ('metadata'    in device &&
           'event_hooks' in device.metadata))
        return ret_enabled;

    for (let event of device.metadata.event_hooks) {
        if (! check_event_trigger(event) ||
            ! check_event_disable(event, indexes.control) ||
            ! check_event_condition(event))
            continue;

        const trigger   = event.trigger;
        const disable   = event.disable;
        const condition = event.condition;

        if (! disable[indexes.control])
            continue;

        let devices = interfaceData.map_invoke[trigger[0]][trigger[1]][trigger[2]];
        for (let dev of devices) {
            let res = condition_check(condition,
                                      interfaceData.devices[dev.databaseId]
                                                   .controls[dev.control]
                                                   .variableInputs[dev.input]
                                                   .properties
                                                   .value);
            if (res)
                return {
                    flag: true,
                    texts: event.texts,
                };
        }

    }

    return ret_enabled;
}



function check_disabled_frontend(uiElement, sibling_idx, sibling_props) {
    const elem_id = uiElement.uniqueUiElementId;
    const ret_enabled = {flag: false};

    if (interfaceData.disable_hooks === undefined ||
        interfaceData.disable_hooks[elem_id] === undefined)
        return ret_enabled;

    const conditions_raw = interfaceData.disable_hooks[elem_id];
    const conditions = Array.isArray(conditions_raw)
                        ? conditions_raw
                        : [conditions_raw];

    for (const condition of conditions) {
        if (condition.condition !== undefined &&
            condition_check(condition.condition, sibling_props) &&
            condition.disable.indexOf(sibling_idx) !== -1)
            return {
                flag: true,
                texts: {
                    title: condition.reason ? condition.reason : '',
                },
            };
    }

    return ret_enabled;
}



Vue.use({
    install: function (Vue, _opts) {
        Vue.prototype.$homegear = homegear;
    },
});



Vue.mixin({
    filters: {
        pretty: function (val) {
            return JSON.stringify(val, null, 4);
        },

        warn: function (val) {
            console.warn(val);
        },

        log: function (val) {
            console.log(val);
        },
    },

    data: function () {
        return  {
            interfaceIcons: icons,
            interfaceData:  interfaceData,
        };
    },

    methods: {
        level3: function (device, _name) {
            function first_in_or(obj, or) {
                if (obj === undefined ||
                    Array.isArray(obj) === false ||
                    obj.length === 0)
                    return or;
                return obj[0];
            }

            const matched = this.$route.matched.map(x => x.name);
            const last    = matched[matched.length - 1];

            if (matched.indexOf('house.tab.rooms.room') !== -1) {
                return this.$router.push({
                    name: `${last}.device`,
                    params: {
                        floor_id: this.floor_id,
                        room_id:  this.room_id,
                        device_id: device,
                    },
                });
            }

            if (matched.indexOf('house.tab.devices') !== -1) {
                const dev = interfaceData.devices[device];
                // TODO: verify that there is always at least a room and a floor
                return this.$router.push({
                    name: `${last}.device`,
                    params: {
                        role_id:   this.role_id,
                        floor_id:  first_in_or(dev.floors, -1),
                        room_id:   first_in_or(dev.rooms, -1),
                        device_id: device,
                    },
                });
            }

            if (matched.indexOf('favorites.list') !== -1) {
                const dev = interfaceData.devices[device];
                // TODO: verify that there is always at least a room and a floor
                return this.$router.push({
                    name: 'favorites.device',
                    params: {
                        floor_id: first_in_or(dev.floors, -1),
                        room_id:  first_in_or(dev.rooms, -1),
                        device_id: device,
                    },
                });
            }

            /**
             * Only act in those two cases!
             * There might have been a click on the disabled content
             * (content_small) which we do not care about.
             **/
            // throw 'BUG: you should not get here';
        },

        round: function (val, precision) {
            const mul = Math.pow(10, precision || 0);

            return Math.round(val * mul) / mul;
        },

        float_formatted: function (val, precision=1) {
            return parseFloat(val).toFixed(precision);
        },

        alert: window.alert,

        i18n: i18n,

        date: date_format,
    },
});



let ShifLogoff = Vue.component('shif-logoff', {
    mounted: function () {
        user_logoff();
    }
});


const scroll_positions = {};


let router = new VueRouter({
    routes: [
        { path: '/',   name: 'index',  redirect: {name: interfaceData.options.startPath || 'house'}, },
        { path: '/nb', name: 'legacy', redirect: {name: 'index'}},

        { path: '/house', name: 'house', component: ShifHouse, redirect: {name: 'house.tab.rooms'},
            children: [
                {
                    name: 'house.tab.rooms',
                    path: 'rooms',
                    component: ShifHouseRooms,
                    meta: {breadcrumbs: ['house', 'house.tab.rooms'], base: true,},
                },
                {
                    name: 'house.tab.rooms.room',
                    path: 'rooms/floor/:floor_id/room/:room_id',
                    components: {small: ShifHouseRooms, big: ShifHouseLvl2},
                    meta: {
                        breadcrumbs: ['house', 'house.tab.rooms', 'house.tab.rooms.room'],
                        cache_ident: {big: {params: ['room_id']}}
                    },
                    props: {small: false, big: true},
                },
                {
                    name: 'house.tab.rooms.room.device',
                    path: 'rooms/floor/:floor_id/room/:room_id/device/:device_id',
                    components: {small: ShifHouseLvl2, big: ShifHouseLvl3},
                    meta: {
                        breadcrumbs: ['house', 'house.tab.rooms', 'house.tab.rooms.room', 'house.tab.rooms.room.device'],
                        cache_ident: {small: {params: ['room_id']}, big: {params: ['room_id', 'device_id']}},
                    },
                    props: {small: true, big: true},
                },

                {
                    name: 'house.tab.devices',
                    path: 'devices',
                    component: ShifHouseDevices,
                    meta: {breadcrumbs: ['house', 'house.tab.devices'], base: true,},
                    props: true,
                },
                {
                    name: 'house.tab.devices.device',
                    path: 'devices/role/:role_id/floor/:floor_id/room/:room_id/device/:device_id',
                    components: {small: ShifHouseDevices, big: ShifAllDevicesLvl3},
                    meta: {
                        breadcrumbs: ['house', 'house.tab.devices', 'house.tab.rooms.room', 'house.tab.devices.device'],
                        cache_ident: {small: {bc_idx: 1}, big: {params: ['role_id', 'room_id', 'device_id']}},
                    },
                    props: {small: true, big: true},
                },

                {
                    name: 'house.tab.profiles',
                    path: 'profiles',
                    component: ShifProfiles,
                    meta: {breadcrumbs: ['house', 'house.tab.profiles'], base: true},
                },
            ],
        },

        { path: '/settings', name: 'settings', component: ShifSettings, redirect: {name: 'settings.list'},
            children: [
                {
                    name: 'settings.list',
                    path: 'list',
                    component: ShifSettingsItems(1),
                    meta: {breadcrumbs: ['settings'], base: true,},
                },
                {
                    name: 'settings.about',
                    path: 'about',
                    components: {small: ShifSettingsItems(1), big: ShifSettingsLicenses},
                    meta: {breadcrumbs: ['settings', 'settings.about']},
                },
                {
                    path: 'user',
                    name: 'settings.user',
                    components: {small: ShifSettingsItems(1), big: ShifSettingsItems(2)},
                    meta: {breadcrumbs: ['settings', 'settings.user']},
                },
                {
                    path: 'user/manage',
                    name: 'settings.user.manage',
                    components: {small: ShifSettingsItems(2), big: ShifSettingsUser},
                    meta: {breadcrumbs: ['settings', 'settings.user', 'settings.user.manage']},
                },
                {
                    path: 'favorites',
                    name: 'settings.favorites',
                    components: {small: ShifSettingsItems(1), big: ShifSettingsFavorites},
                    meta: {breadcrumbs: ['settings', 'settings.favorites']}
                },
                {
                    path: 'sort',
                    name: 'settings.sort',
                    components: {small: ShifSettingsItems(1), big: ShifSettingsSort},
                    meta: {breadcrumbs: ['settings', 'settings.sort']}
                },
                {
                    path: 'profiles',
                    name: 'settings.profiles',
                    components: {small: ShifSettingsItems(1), big: ShifSettingsProfiles},
                    meta: {breadcrumbs: ['settings', 'settings.profiles']}
                }, {
                    path: 'profiles/add',
                    name: 'settings.profiles.new',
                    components: {small: ShifSettingsProfiles, big: ShifSettingsProfile},
                    meta: {breadcrumbs: ['settings', 'settings.profiles', 'settings.profiles.new']}
                }, {
                    path: 'profiles/edit/:profile_id',
                    name: 'settings.profiles.profile',
                    components: {small: ShifSettingsProfiles, big: ShifSettingsProfile},
                    meta: {breadcrumbs: ['settings', 'settings.profiles', 'settings.profiles.profile']},
                    props: {small: false, big: true},
                },
                {
                    path: 'automations',
                    name: 'settings.automations',
                    components: {small: ShifSettingsItems(1), big: ShifSettingsAutomations},
                    meta: {breadcrumbs: ['functions', 'settings.automations'],},
                },
                {
                    path: 'automations/selection/:automation_ids',
                    name: 'settings.automations.selection',
                    components: {small: ShifSettingsItems(1), big: ShifSettingsAutomations},
                    meta: {
                        breadcrumbs: ['functions', 'settings.automations', 'settings.automations.selection'],
                        cache_ident: {big: {params: ['automation_ids']}},
                    },
                    props: {small: false, big: true},
                },
                {
                    path: 'automations/add',
                    name: 'settings.automations.new',
                    components: {small: ShifSettingsAutomations, big: ShifSettingsAutomationsForm},
                    meta: {breadcrumbs: ['settings', 'settings.automations', 'settings.automations.new'],},
                },
                {
                    path: 'automations/edit/:automation_id',
                    name: 'settings.automations.automation',
                    components: {small: ShifSettingsAutomations, big: ShifSettingsAutomationsForm},
                    meta: {
                        breadcrumbs: ['settings', 'settings.automations', 'settings.automations.automation'],
                        cache_ident: {big: {params: ['automation_id']}},
                    },
                    props: {small: false, big: true},
                },
            ],
        },

        { path: '/favorites', name: 'favorites', component: ShifFavorites, redirect: {name: 'favorites.list'},
            children: [
                {
                    name: 'favorites.list',
                    path: 'list',
                    component: ShifFavoritesLvl1,
                    meta: {breadcrumbs: ['favorites'], base: true},
                },
                {
                    name: 'favorites.device',
                    path: 'floor/:floor_id/room/:room_id/device/:device_id',
                    components: {small: ShifFavoritesLvl1, big: ShifFavoritesLvl3},
                    meta: {
                        breadcrumbs: ['favorites', 'house.tab.rooms.room', 'favorites.device'],
                        cache_ident: {small: {bc_idx: 0}, big: {params: ['room_id', 'device_id']}}
                    },
                    props: {small: false, big: true},
                }
            ],
        },

        { path: '/notifications', name: 'notifications', component: ShifNotifications, redirect: {name: 'notifications.list'},
            children: [
                {
                    name: 'notifications.list',
                    path: 'list',
                    component: ShifNotificationsLvl1,
                    meta: {breadcrumbs: ['notifications'], base: true},
                },
                {
                    name: 'notifications.notification',
                    path: ':notification_id',
                    components: {small: ShifNotificationsLvl1, big: ShifNotificationsNotification},
                    meta: {
                        breadcrumbs: ['notifications', 'notifications.notification'],
                        cache_ident: {big: {params: ['notification_id']}},
                    },
                    props: {small: false, big: true},
                },
            ],
        },

        {
            path: '/log',
            name: 'log',
            component: ShifLog,
            meta: {breadcrumbs: ['log']},
        },

        { path: '/logoff',    name: 'logoff',    component: ShifLogoff, },
    ],
});



let app = new Vue({
    name: 'App',

    mixins: [mixin_modemenu],

    components: {
        ShifMainmenu,
        ShifModemenu,
    },

    data: {
        debug: false,

        favorites: {
            enabled: false,
        },
        profiles: {
            devs:    {},
            enabled: false,
            active:  null,
        },
        draggable: {
            enabled: false,
            in_progress: false,
        },

        show: false,
    },

    router: router,

    // Hack: decrease .content height when modemenu is enabled.
    template: `
        <div v-if="show"
             id="inhalt"
             v-bind:class="{'modemenu-visible': modemenu_is_shown}">
            <router-view />

            <shif-modemenu />
            <shif-mainmenu />
        </div>
    `,
});



let breadcrumbs = new Vue({
    name: 'Breadcrumbs',

    router: router,

    data: {
        show: false,
    },

    computed:  {
        routes_with_proper_names: function () {
            if (this.$route.meta === undefined ||
                this.$route.meta.breadcrumbs === undefined)
                return [];

            let routes = this.$route.meta.breadcrumbs
                                         .map(name => ({
                                            link:     name,
                                            name:     this.get_name(name),
                                            disabled: false,
                                         }))
                                         .filter(x => x.name !== '?');

            if (routes.length > 0)
                routes[routes.length - 1].disabled = true;

            return routes;
        },

        back_wanted: function () {
            return (! this.$route.meta.base) &&
                   this.routes_with_proper_names.length > 1;
        },
    },

    methods: {
        get_name: function (route_name) {
            function floor() {
                if (! interfaceData.options.showFloor)
                    return '';

                return Number(params.floor_id) === -1
                    ? i18n('house.storyless') + ' - '
                    : interfaceData.floors[params.floor_id].name + ' - ';
            }

            function room() {
                return Number(params.room_id) === -1
                    ? i18n('house.storyless')
                    : interfaceData.rooms[params.room_id].name;
            }

            const params  = this.$route.params;

            switch (route_name) {
                case 'house.tab.rooms.room':
                    return floor() + room();

                case 'house.tab.rooms.room.device':
                case 'house.tab.devices.device':
                case 'favorites.device':
                    if (interfaceData.devices[params.device_id] !== undefined)
                        return interfaceData.devices[params.device_id].label;
                    break;

                case 'log':
                    return 'Log';

                case 'settings.profiles.profile':
                    if (interfaceData.profiles[params.profiles_id] !== undefined)
                        return interfaceData.profiles[params.profile_id].name;
                    break;

                case 'settings.automations.automation':
                    if (interfaceData.automations[params.automation_id] !== undefined)
                        return interfaceData.automations[params.automation_id].name;
                    break;

                case 'notifications.notification':
                    if (interfaceData.notifications[params.notification_id] !== undefined)
                        return interfaceData.notifications[params.notification_id].title;
                    break;
            }

            return i18n(route_name);
        },
    },

    template: `
        <div v-if="show"
             id="breadcrumbs_wrapper">
            <div id="breadcrumbs">
                <template>
                    <shif-icon id="back" src="arrow_left_1"
                               v-bind:style="{visibility: back_wanted ? 'visible' : 'hidden'}"
                               v-on:click="$router.back()"/>
                </template>
                <div id="breadcrumb_wrapper">
                    <template v-for="i in routes_with_proper_names">
                        <router-link v-bind:to="{name: i.link}"
                                     v-bind:disabled="i.disabled"
                                     v-bind:class="{disabled: i.disabled}"
                                     >{{ i.name }}</router-link>
                    </template>
                </div>
            </div>
        </div>
    `
});



let toast = new Vue({
    el: '#toast',

    name: 'Toast',

    router: router,

    mixins: [mixin_notification],

    data: {
        msgs: [],
    },

    computed: {
        msgs_combined: function () {
            return this.msgs.concat(this.toast_notifications);
        },
    },

    methods: {
        remove_msg: function (msg_idx) {
            this.msgs.splice(msg_idx, 1);
        },

        push: function (msg) {
            this.msgs.unshift(msg);
        },

        set: function (msg) {
            this.msgs = [msg];
        },

        link: function (id) {
            return {
                name: 'notifications.notification',
                params: {
                    notification_id: id,
                },
            };
        },

        is_toast_integrated: function (msg) {
            return msg.displayType === NotificationDisplayType.ToastIntegrated;
        },

        click: function (msg, btn) {
            if (msg.id !== undefined)
                this.notification_action(msg.id, btn.id, () => {
                    if (btn.reloadUi === true)
                        window.location.reload();
                });
            else if (btn.reloadUi === true)
                window.location.reload();
        },

        notification_hide_or_remove: function (msg) {
            if (msg.buttons.length === 0)
                this.notification_remove(msg.id);
            else
                this.notification_hide(msg.id);
        },
    },

    /**
     * msg.close is only available in notifications not handled by homegear
     * msg.flags.closeable is only available in notifications handled by homegear
     **/
    template: `
        <div id="toast">
            <div class="toast" v-for="msg, i in msgs_combined"
                               v-bind:key="msg.id">
                    <template v-if="msg.close">
                        <button class="toast_close"
                                v-on:click="remove_msg(i)">
                            X
                        </button>
                    </template>
                    <template v-if="msg.flags && msg.flags.closeable === true">
                        <button class="toast_close"
                                v-on:click="notification_hide_or_remove(msg)">
                            X
                        </button>
                    </template>

                    <template v-if="is_toast_integrated(msg)">
                        <router-link v-bind:to="link(msg.id)"
                                     class="toast_content" v-html="msg.content">
                        </router-link>
                    </template>
                    <template v-else>
                        <div class="toast_content">
                            <div class="toast_text" v-html="msg.content">
                            </div>
                            <template v-for="btn in msg.buttons">
                                <button class="toast_action"
                                        v-bind:class="btn.type"
                                        v-on:click="click(msg, btn)">
                                    {{ btn.label }}
                                </button>
                            </template>
                        </div>
                    </template>
            </div>
        </div>
    `
});
                    // <div class="toast_content" v-html="msg.content"></div>



function interface_mount(show=true) {
    app.$mount('#inhalt');
    breadcrumbs.$mount('#breadcrumbs');

    if (show)
        interface_show();
}



function interface_show() {
    app.show = true;
    breadcrumbs.show = true;
}



/**
 * type: 'requestUiRefresh' || any
 * buttons: bool
 **/
function _create_notification(inc_buttons = false, args) {
    const buttons = ! inc_buttons ? [] : [
        {
            id: 0,
            type: 'success',
            reloadUi: false,
            closeModal: false,
            label: 'Yes I`ve read that',
            icon: 'abort_1',
        },
        {
            id: 1,
            type: 'warning',
            reloadUi: false,
            closeModal: true,
            label: 'ignore',
            icon: 'abort_1',
        },
        {
            id: 2,
            type: 'error',
            reloadUi: true,
            closeModal: false,
            label: 'reload',
            icon: 'abort_1',
        },
    ];

    const msg = {
        type: 'test',
        flags: {
            closeable: args.closeable || false,
            hasModal: args.hasModal || false,
            overlayNotification: args.overlayNotification || false,
            overlayModal: args.overlayModal || false,
        },
        title: {
            'de-DE': 'Test Nachricht: ' + date_format(),
            'en-US': 'Test Message: ' + date_format(),
        },
        content: `
            <p class="toast_text">Test</p>
        `,
        modalTitle: {
            'de-DE': 'Test Nachricht: Modal',
            'en-US': 'Test Message: Modal',
        },
        modalContent: {
            'de-DE': 'Test Content: ' + date_format(),
            'en-US': 'Test Content: ' + date_format(),
        },
        buttons,
    };

    homegear.invoke({
        jsonrpc: '2.0',
        method: 'createUiNotification',
        params: [msg],
    }, function () {
        homegear.invoke({jsonrpc: '2.0', method: 'getUiNotifications', params: ['de-DE']}, console.log);
        homegear.invoke({jsonrpc: '2.0', method: 'getUiNotifications', params: ['en-US']}, console.log);
    });
}



function _cleanup_notifications() {
    homegear.invoke({
        jsonrpc: '2.0',
        method: 'getUiNotifications',
        params: [interfaceData.options.language],
    }, function (resp) {
        for (const i of resp.result)
            homegear.invoke({
                jsonrpc: '2.0',
                method: 'removeUiNotification',
                params: [i.id],
            });
    });
}



const _modal = {
    closeable: true,
    hasModal: true,
    overlayModal: true,
    overlayNotification: false,
};

const _toast_only = {
    closeable: true,
    hasModal: false,
    overlayModal: false,
    overlayNotification: true,
};

const _toast_integrated = {
    closeable: true,
    hasModal: true,
    overlayModal: false,
    overlayNotification: true,
};

const _integrated = {
    closeable: false,
    hasModal: true,
    overlayModal: false,
    overlayNotification: false,
};

